from __future__ import annotations

import inspect
from typing import Any, Awaitable, Callable

from ... import events as _ev
from ...stages import Operated, Resolved
from ...types import Atom, Ctx, OperatedCtx

ANCHOR = _ev.SYS_HANDLER_PERSISTENCE

HandlerFn = Callable[[object | None, Any], object | Awaitable[object]]


def _resolve_handler(obj: object | None, ctx: Any) -> HandlerFn | None:
    fn = getattr(ctx, "handler", None)
    if callable(fn):
        return fn

    model = getattr(ctx, "model", None)
    if model is None and obj is not None:
        model = type(obj)

    alias = getattr(ctx, "op", None) or getattr(ctx, "method", None)
    if model is None or not isinstance(alias, str):
        return None

    model_fn = getattr(model, alias, None)
    if callable(model_fn):
        return lambda _obj, _ctx: model_fn(_ctx)

    return None


class BoundAtom(Atom[Resolved, Operated]):
    name = "handler.custom"
    anchor = ANCHOR

    def __init__(self, fn: HandlerFn, *, name: str | None = None) -> None:
        self._fn = fn
        if name:
            self.name = name

    async def __call__(self, obj: object | None, ctx: Ctx[Resolved]) -> Ctx[Operated]:
        result = self._fn(obj, ctx)
        if inspect.isawaitable(result):
            result = await result
        setattr(ctx, "result", result)
        return ctx.promote(OperatedCtx)


def bind(fn: HandlerFn, *, name: str | None = None) -> Atom[Resolved, Operated]:
    return BoundAtom(fn, name=name)


class AtomImpl(Atom[Resolved, Operated]):
    name = "handler.custom"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Resolved]) -> Ctx[Operated]:
        fn = _resolve_handler(obj, ctx)
        if fn is None:
            return ctx.promote(OperatedCtx)
        result = fn(obj, ctx)
        if inspect.isawaitable(result):
            result = await result
        setattr(ctx, "result", result)
        return ctx.promote(OperatedCtx)


INSTANCE = AtomImpl()


__all__ = ["ANCHOR", "HandlerFn", "BoundAtom", "bind", "INSTANCE"]



from __future__ import annotations

from typing import Any

import tigrbl_ops_oltp as _core

from ... import events as _ev
from ...stages import Operated, Resolved
from ...types import Atom, Ctx, OperatedCtx
from . import _oltp_context as _ctx

ANCHOR = _ev.SYS_HANDLER_PERSISTENCE


async def _run(obj: object | None, ctx: Any) -> None:
    model = obj if isinstance(obj, type) else getattr(ctx, "model", None)
    if not isinstance(model, type):
        raise TypeError("handler_bulk_merge requires a model type")
    payload = _ctx.payload(ctx)
    if not isinstance(payload, list):
        raise TypeError("bulk_merge expects a list payload")
    setattr(ctx, "result", await _core.bulk_merge(model, payload, db=_ctx.db(ctx)))


class AtomImpl(Atom[Resolved, Operated]):
    name = "sys.handler_bulk_merge"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Resolved]) -> Ctx[Operated]:
        await _run(obj, ctx)
        return ctx.promote(OperatedCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



from __future__ import annotations

from ...types import Atom, Ctx, OperatedCtx
from ...stages import Operated

from typing import Any, Optional
import logging

from ... import events as _ev
from ..._opview_helpers import _ensure_schema_out

# Runs late in POST_HANDLER, before out model build and dumping.
ANCHOR = _ev.SCHEMA_COLLECT_OUT  # "schema:collect_out"

logger = logging.getLogger("uvicorn")


def _run(obj: Optional[object], ctx: Any) -> None:
    """Load precompiled outbound schema into ctx.temp."""
    del obj
    temp = getattr(ctx, "temp", None)
    if not isinstance(temp, dict):
        temp = {}
        setattr(ctx, "temp", temp)

    if isinstance(temp.get("schema_out"), dict):
        return

    temp["schema_out"] = dict(_ensure_schema_out(ctx))


class AtomImpl(Atom[Operated, Operated]):
    name = "schema.collect_out"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Operated]) -> Ctx[Operated]:
        _run(obj, ctx)
        return ctx.promote(OperatedCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



from __future__ import annotations

from ...types import Atom, Ctx, EmittingCtx
from ...stages import Encoded, Emitting

import json
from typing import Any, Mapping

from ... import events as _ev
from .._temp import _ensure_temp

ANCHOR = _ev.EGRESS_TO_TRANSPORT_RESPONSE


def _jsonrpc_request_id(ctx: Any) -> Any:
    temp = getattr(ctx, "temp", None)
    route = temp.get("route", {}) if isinstance(temp, dict) else {}
    rpc_env = route.get("rpc_envelope") if isinstance(route, dict) else None
    if isinstance(rpc_env, Mapping):
        return rpc_env.get("id")

    raw = getattr(ctx, "gw_raw", None) or getattr(ctx, "raw", None)
    rpc = getattr(raw, "rpc", None)
    if isinstance(rpc, Mapping):
        return rpc.get("id")

    temp = getattr(ctx, "temp", None)
    dispatch = temp.get("dispatch", {}) if isinstance(temp, dict) else {}
    if isinstance(dispatch, Mapping):
        rpc_dispatch = dispatch.get("rpc")
        if isinstance(rpc_dispatch, Mapping):
            return rpc_dispatch.get("id")

    body = getattr(ctx, "body", None)
    if isinstance(body, (bytes, bytearray)):
        try:
            body = json.loads(bytes(body).decode("utf-8"))
        except Exception:
            body = None
    if isinstance(body, Mapping) and body.get("jsonrpc") == "2.0":
        return body.get("id")

    if isinstance(dispatch, Mapping):
        parsed_payload = dispatch.get("parsed_payload")
        if isinstance(parsed_payload, Mapping) and "id" in parsed_payload:
            return parsed_payload.get("id")

    request = getattr(ctx, "request", None)
    req_body = getattr(request, "_json", None)
    if isinstance(req_body, Mapping) and req_body.get("jsonrpc") == "2.0":
        return req_body.get("id")
    return None


def _is_jsonrpc_request(ctx: Any) -> bool:
    raw = getattr(ctx, "gw_raw", None) or getattr(ctx, "raw", None)
    if getattr(raw, "kind", None) == "jsonrpc":
        return True

    temp = getattr(ctx, "temp", None)
    route = temp.get("route", {}) if isinstance(temp, dict) else {}
    rpc_env = route.get("rpc_envelope") if isinstance(route, dict) else None
    if isinstance(rpc_env, Mapping) and rpc_env.get("jsonrpc") == "2.0":
        return True

    path = getattr(raw, "path", None)
    app = getattr(ctx, "app", None)
    prefix = getattr(app, "jsonrpc_prefix", None)
    if isinstance(path, str) and isinstance(prefix, str):
        return (path.rstrip("/") or "/") == (prefix.rstrip("/") or "/")

    return False


def _normalize_jsonrpc_transport_response(
    ctx: Any, response: dict[str, Any]
) -> dict[str, Any]:
    normalized = dict(response)
    request_id = _jsonrpc_request_id(ctx)
    if request_id is None:
        normalized["status_code"] = 204
        normalized["body"] = b""
        return normalized

    body = normalized.get("body")
    if body is None:
        body = getattr(ctx, "result", None)
    if isinstance(body, (bytes, bytearray)):
        try:
            body = json.loads(bytes(body).decode("utf-8"))
        except Exception:
            body = bytes(body).decode("utf-8", errors="replace")

    if not (isinstance(body, Mapping) and body.get("jsonrpc") == "2.0"):
        status_code = int(normalized.get("status_code", 200) or 200)
        if status_code >= 400:
            from tigrbl_typing.status.mappings import ERROR_MESSAGES, _HTTP_TO_RPC

            rpc_code = _HTTP_TO_RPC.get(status_code, -32603)
            message = ERROR_MESSAGES.get(rpc_code, "Internal error")
            if isinstance(body, Mapping):
                detail = body.get("detail")
                if isinstance(detail, str) and detail:
                    message = detail
                data = dict(body)
            else:
                data = {"detail": body}
            body = {
                "jsonrpc": "2.0",
                "error": {"code": rpc_code, "message": message, "data": data},
                "id": request_id,
            }
        else:
            body = {"jsonrpc": "2.0", "result": body, "id": request_id}

    normalized["status_code"] = 200
    normalized["body"] = body
    return normalized


def _run(obj: object | None, ctx: Any) -> None:
    del obj
    temp = _ensure_temp(ctx)
    egress = temp.setdefault("egress", {})

    existing = egress.get("transport_response")
    if isinstance(existing, dict):
        if _is_jsonrpc_request(ctx):
            normalized = _normalize_jsonrpc_transport_response(ctx, existing)
            egress["transport_response"] = normalized
            setattr(ctx, "transport_response", normalized)
        return

    body = egress.get("enveloped")
    if body is None:
        body = egress.get("wire_payload")
    if body is None:
        body = getattr(ctx, "result", None)
    if body is None:
        response_ns = getattr(ctx, "response", None)
        if response_ns is not None:
            body = getattr(response_ns, "result", None)

    headers_obj = egress.get("headers", getattr(ctx, "response_headers", {})) or {}
    if isinstance(headers_obj, dict):
        headers = dict(headers_obj)
    else:
        data = getattr(headers_obj, "__dict__", None)
        if isinstance(data, dict):
            headers = dict(data)
        else:
            try:
                headers = dict(headers_obj)
            except Exception:
                headers = {}

    status_code = int(
        egress.get("status_code", getattr(ctx, "status_code", 200)) or 200
    )
    response = {
        "status_code": status_code,
        "headers": headers,
        "body": body,
    }
    if _is_jsonrpc_request(ctx):
        response = _normalize_jsonrpc_transport_response(ctx, response)

    egress["transport_response"] = response
    setattr(ctx, "transport_response", response)


class AtomImpl(Atom[Encoded, Emitting]):
    name = "egress.to_transport_response"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Encoded]) -> Ctx[Emitting]:
        _run(obj, ctx)
        return ctx.promote(EmittingCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from ...phases import PhaseName


async def _maybe_await(value: Any) -> Any:
    if hasattr(value, "__await__"):
        return await value
    return value


@dataclass(frozen=True, slots=True)
class DbCapabilities:
    allow_flush: bool
    allow_commit: bool
    commit_requires_owned_tx: bool
    allow_refresh: bool = False


_PHASE_DB_CAPABILITIES: dict[PhaseName, DbCapabilities] = {
    "INGRESS_BEGIN": DbCapabilities(False, False, True, False),
    "INGRESS_PARSE": DbCapabilities(False, False, True, False),
    "INGRESS_DISPATCH": DbCapabilities(False, False, True, False),
    "PRE_TX_BEGIN": DbCapabilities(False, False, True, False),
    "START_TX": DbCapabilities(False, False, True, False),
    "PRE_HANDLER": DbCapabilities(True, False, True, False),
    "HANDLER": DbCapabilities(True, True, True, False),
    "POST_HANDLER": DbCapabilities(True, False, True, False),
    "PRE_COMMIT": DbCapabilities(False, False, True, False),
    "END_TX": DbCapabilities(True, True, True, False),
    "POST_COMMIT": DbCapabilities(True, False, True, True),
    "POST_RESPONSE": DbCapabilities(False, False, True, False),
    "EGRESS_SHAPE": DbCapabilities(False, False, True, False),
    "EGRESS_FINALIZE": DbCapabilities(False, False, True, False),
    "ON_ERROR": DbCapabilities(False, False, True, False),
    "ON_PRE_TX_BEGIN_ERROR": DbCapabilities(False, False, True, False),
    "ON_START_TX_ERROR": DbCapabilities(False, False, True, False),
    "ON_PRE_HANDLER_ERROR": DbCapabilities(False, False, True, False),
    "ON_HANDLER_ERROR": DbCapabilities(False, False, True, False),
    "ON_POST_HANDLER_ERROR": DbCapabilities(False, False, True, False),
    "ON_PRE_COMMIT_ERROR": DbCapabilities(False, False, True, False),
    "ON_END_TX_ERROR": DbCapabilities(False, False, True, False),
    "ON_POST_COMMIT_ERROR": DbCapabilities(False, False, True, False),
    "ON_POST_RESPONSE_ERROR": DbCapabilities(False, False, True, False),
    "ON_ROLLBACK": DbCapabilities(False, False, True, False),
}


def phase_db_capabilities(phase: str) -> DbCapabilities:
    try:
        return _PHASE_DB_CAPABILITIES[phase]  # type: ignore[index]
    except KeyError as exc:  # pragma: no cover
        raise RuntimeError(f"Unknown phase for PhaseDb binding: {phase!r}") from exc


class PhaseDb:
    __slots__ = ("_db", "_phase", "_caps", "_owns_tx")

    def __init__(
        self,
        db: Any,
        *,
        phase: str,
        caps: DbCapabilities,
        owns_tx: bool,
    ) -> None:
        self._db = db
        self._phase = phase
        self._caps = caps
        self._owns_tx = owns_tx

    @property
    def raw(self) -> Any:
        return self._db

    def _deny(self, op: str) -> RuntimeError:
        return RuntimeError(f"db.{op}() is not allowed during {self._phase} phase")

    async def flush(self) -> None:
        if not self._caps.allow_flush:
            raise self._deny("flush")
        await _maybe_await(self._db.flush())

    async def commit(self) -> None:
        if not self._caps.allow_commit:
            raise self._deny("commit")
        if self._caps.commit_requires_owned_tx and not self._owns_tx:
            raise self._deny("commit")
        await _maybe_await(self._db.commit())

    async def refresh(self, instance: Any) -> None:
        if not self._caps.allow_refresh:
            raise self._deny("refresh")
        await _maybe_await(self._db.refresh(instance))

    async def rollback(self) -> None:
        await _maybe_await(self._db.rollback())

    def __getattr__(self, name: str) -> Any:
        return getattr(self._db, name)


def bind_phase_db(ctx: Any) -> Any:
    raw_db = getattr(ctx, "_raw_db", None)
    if raw_db is None:
        ctx.db = None
        return ctx

    phase = getattr(ctx, "phase", None)
    if not isinstance(phase, str) or not phase:
        raise RuntimeError("ctx.phase must be set before PhaseDb binding")

    owns_tx = bool(getattr(ctx, "owns_tx", False))
    temp = getattr(ctx, "temp", None)
    cache_token = (id(raw_db), owns_tx)
    cache: dict[str, PhaseDb] | None = None
    if isinstance(temp, dict):
        cached_token = temp.get("_phase_db_cache_token")
        if cached_token == cache_token:
            maybe_cache = temp.get("_phase_db_cache")
            if isinstance(maybe_cache, dict):
                cache = maybe_cache
        else:
            temp["_phase_db_cache_token"] = cache_token
            temp["_phase_db_cache"] = {}
            cache = temp["_phase_db_cache"]

    if cache is not None:
        wrapped = cache.get(phase)
        if wrapped is None:
            wrapped = PhaseDb(
                raw_db,
                phase=phase,
                caps=phase_db_capabilities(phase),
                owns_tx=owns_tx,
            )
            cache[phase] = wrapped
        ctx.db = wrapped
        return ctx

    ctx.db = PhaseDb(
        raw_db,
        phase=phase,
        caps=phase_db_capabilities(phase),
        owns_tx=owns_tx,
    )
    return ctx


def run(ctx: Any) -> Any:
    return bind_phase_db(ctx)


__all__ = ["DbCapabilities", "PhaseDb", "bind_phase_db", "phase_db_capabilities", "run"]



# pkgs/standards/tigrbl_atoms/tigrbl/atoms/refresh/demand.py
from __future__ import annotations

from ...types import Atom, Ctx, OperatedCtx
from ...stages import Operated

from typing import Any, Iterable, Optional
import logging

from ... import events as _ev
from ..._opview_helpers import _ensure_ov, _ensure_temp

# After the handler flushes changes; decide whether to hydrate DB-generated values.
ANCHOR = _ev.POST_FLUSH  # "post:flush"

logger = logging.getLogger("uvicorn")


def _run(obj: Optional[object], ctx: Any) -> None:
    """
    refresh:demand@post:flush

    Purpose
    -------
    Decide whether to hydrate refreshed values from the DB after a write (INSERT/UPDATE).
    We do NOT perform the refresh here; we only mark intent on the context so the
    executor (or handler) can perform the vendor-specific action (RETURNING/refresh()).

    Inputs (conventions)
    --------------------
    - ctx.persist : bool               → write op? (anchor is persist-tied but we guard)
    - ctx.temp["used_returning"] : bool              → a prior step already satisfied hydration
    - ctx.temp["hydrated_values"] : Mapping[str, Any]→ values captured from RETURNING
    - ctx.cfg.refresh_after_write : Optional[bool]   → policy override (true/false)
      (also checks ctx.cfg.refresh_policy.{always,never,auto})
    - ctx.bulk    : Optional[bool]    → bulk write; executor may choose a different strategy

    Effects
    -------
    - ctx.temp["refresh_demand"] : bool
    - ctx.temp["refresh_fields"] : tuple[str, ...] (hint: which fields likely changed in DB)
    - ctx.temp["refresh_reason"] : str (diagnostic only)
    """
    logger.debug("Running refresh:demand")
    if getattr(ctx, "persist", True) is False:
        logger.debug("Skipping refresh:demand; ctx.persist is False")
        return

    temp = _ensure_temp(ctx)
    ov = _ensure_ov(ctx)
    refresh_hints = tuple(ov.refresh_hints)

    # If RETURNING already produced hydrated values, skip unless policy forces refresh.
    returning_satisfied = bool(temp.get("used_returning")) or bool(
        temp.get("hydrated_values")
    )
    logger.debug("Returning satisfied: %s", returning_satisfied)

    # Policy: cfg.refresh_after_write wins if explicitly set; otherwise "auto".
    policy = _get_refresh_policy(ctx)
    logger.debug("Refresh policy: %s", policy)
    # auto → infer from specs (db-generated signals) OR absence of returning values
    needs_by_specs = bool(refresh_hints)
    logger.debug("Refresh hints: %s; fields=%s", needs_by_specs, refresh_hints)
    need_refresh = _decide(policy, returning_satisfied, needs_by_specs)
    logger.debug("Refresh decision: %s", need_refresh)

    temp["refresh_demand"] = bool(need_refresh)
    temp["refresh_fields"] = refresh_hints

    if need_refresh:
        temp["refresh_reason"] = _reason(
            policy, returning_satisfied, needs_by_specs, refresh_hints
        )
        logger.debug("Refresh scheduled: %s", temp["refresh_reason"])
    else:
        temp["refresh_reason"] = "skipped: returning_satisfied or policy=false"
        logger.debug("Refresh skipped: %s", temp["refresh_reason"])

    # Executor/handler will look at ctx.temp["refresh_demand"] and act accordingly.


# ──────────────────────────────────────────────────────────────────────────────
# Internals
# ──────────────────────────────────────────────────────────────────────────────


def _get_refresh_policy(ctx: Any) -> str:
    """
    Return 'always' | 'never' | 'auto'.
    Accepts any of the following on ctx.cfg (first hit wins):
      - cfg.refresh_after_write: bool → True:'always' / False:'never'
      - cfg.refresh_policy: str in {'always','never','auto'}
    Defaults to 'auto'.
    """
    cfg = getattr(ctx, "cfg", None)
    if cfg is not None:
        val = getattr(cfg, "refresh_after_write", None)
        if isinstance(val, bool):
            return "always" if val else "never"
        pol = getattr(getattr(cfg, "refresh_policy", None), "value", None) or getattr(
            cfg, "refresh_policy", None
        )
        if isinstance(pol, str) and pol in {"always", "never", "auto"}:
            return pol
    return "auto"


def _decide(policy: str, returning_satisfied: bool, needs_by_specs: bool) -> bool:
    if policy == "always":
        return True
    if policy == "never":
        return False
    # auto
    if returning_satisfied:
        # RETURNING already hydrated values; only refresh if specs strongly indicate more work.
        return bool(needs_by_specs)
    # No returning: default to refresh to honor "hydrate after flush" decision.
    return True


def _reason(
    policy: str, returning_satisfied: bool, needs_by_specs: bool, fields: Iterable[str]
) -> str:
    parts = [f"policy={policy}"]
    parts.append(f"returning_satisfied={bool(returning_satisfied)}")
    parts.append(f"specs_need={bool(needs_by_specs)}")
    if fields:
        parts.append(f"fields={','.join(fields)}")
    return "; ".join(parts)


class AtomImpl(Atom[Operated, Operated]):
    name = "refresh.demand"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Operated]) -> Ctx[Operated]:
        _run(obj, ctx)
        return ctx.promote(OperatedCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



from __future__ import annotations

from typing import Any, Callable, Dict, Optional, Tuple

from . import extra as _extra
from . import security as _security

RunFn = Callable[[Optional[object], Any], Any]

REGISTRY: Dict[Tuple[str, str], Tuple[str, RunFn]] = {
    ("dep", "security"): (_security.ANCHOR, _security.INSTANCE),
    ("dep", "extra"): (_extra.ANCHOR, _extra.INSTANCE),
}

__all__ = ["REGISTRY", "RunFn"]



from __future__ import annotations

from typing import Any, Callable, Dict, Optional, Tuple

from ... import events as _ev
from . import asgi_send as _asgi_send
from . import envelope_apply as _envelope_apply
from . import headers_apply as _headers_apply
from . import http_finalize as _http_finalize
from . import out_dump as _out_dump
from . import result_normalize as _result_normalize
from . import to_transport_response as _to_transport_response

RunFn = Callable[[Optional[object], Any], Any]

_ORDERED: Tuple[Tuple[str, str, str, RunFn], ...] = (
    (
        "egress",
        "result_normalize",
        _result_normalize.ANCHOR,
        _result_normalize.INSTANCE,
    ),
    ("egress", "out_dump", _out_dump.ANCHOR, _out_dump.INSTANCE),
    ("egress", "envelope_apply", _envelope_apply.ANCHOR, _envelope_apply.INSTANCE),
    ("egress", "headers_apply", _headers_apply.ANCHOR, _headers_apply.INSTANCE),
    ("egress", "http_finalize", _http_finalize.ANCHOR, _http_finalize.INSTANCE),
    (
        "egress",
        "to_transport_response",
        _to_transport_response.ANCHOR,
        _to_transport_response.INSTANCE,
    ),
    ("egress", "asgi_send", _asgi_send.ANCHOR, _asgi_send.INSTANCE),
)

REGISTRY: Dict[Tuple[str, str], Tuple[str, RunFn]] = {
    (domain, subject): (anchor, run) for domain, subject, anchor, run in _ORDERED
}

_EVENT_ORDER = {name: idx for idx, name in enumerate(_ev.all_events_ordered())}
if tuple(REGISTRY):
    _anchors = [anchor for anchor, _run in REGISTRY.values()]
    if _anchors != sorted(_anchors, key=lambda a: _EVENT_ORDER.get(a, 10_000)):
        raise RuntimeError(
            "Egress atom registry order must follow runtime event order."
        )

__all__ = ["REGISTRY", "RunFn"]



from __future__ import annotations

from typing import Any

from ...types import Atom, Ctx, OperatedCtx
from ...stages import Operated

from ... import events as _ev
from .._temp import _ensure_temp
from ._db import _in_transaction, _resolve_db_handle

ANCHOR = _ev.SYS_TX_COMMIT


async def _run(obj: object | None, ctx: Any) -> None:
    del obj
    temp = _ensure_temp(ctx)
    db = _resolve_db_handle(ctx)
    open_flag = bool(temp.get("__sys_tx_open__")) or (
        db is not None and _in_transaction(db)
    )

    try:
        if not open_flag or db is None:
            return
        commit = getattr(db, "commit", None)
        if callable(commit):
            rv = commit()
            if hasattr(rv, "__await__"):
                await rv
    finally:
        temp["__sys_tx_open__"] = False
        release = temp.pop("__sys_db_release__", None)
        if callable(release):
            release()


class AtomImpl(Atom[Operated, Operated]):
    name = "sys.commit_tx"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Operated]) -> Ctx[Operated]:
        await _run(obj, ctx)
        return ctx.promote(OperatedCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



from __future__ import annotations
from typing import Any, Callable, Dict, Optional, Tuple

from . import template as _template
from . import negotiate as _negotiate
from . import render as _render
from . import error_to_transport as _error_to_transport
from . import headers_from_payload as _hdr_payload

RunFn = Callable[[Optional[object], Any], Any]

REGISTRY: Dict[Tuple[str, str], Tuple[str, RunFn]] = {
    ("response", "template"): (_template.ANCHOR, _template.INSTANCE),
    ("response", "negotiate"): (_negotiate.ANCHOR, _negotiate.INSTANCE),
    ("response", "headers_from_payload"): (_hdr_payload.ANCHOR, _hdr_payload.INSTANCE),
    ("response", "render"): (_render.ANCHOR, _render.INSTANCE),
    ("response", "error_to_transport"): (
        _error_to_transport.ANCHOR,
        _error_to_transport.INSTANCE,
    ),
}

__all__ = ["REGISTRY", "RunFn"]



from __future__ import annotations

from typing import Any, Callable, Dict, Optional, Tuple

from . import binding_match, binding_parse, input_normalize, op_resolve

RunFn = Callable[[Optional[object], Any], Any]

REGISTRY: Dict[Tuple[str, str], Tuple[str, RunFn]] = {
    ("dispatch", "binding_match"): (binding_match.ANCHOR, binding_match.INSTANCE),
    ("dispatch", "binding_parse"): (binding_parse.ANCHOR, binding_parse.INSTANCE),
    ("dispatch", "input_normalize"): (
        input_normalize.ANCHOR,
        input_normalize.INSTANCE,
    ),
    ("dispatch", "op_resolve"): (op_resolve.ANCHOR, op_resolve.INSTANCE),
}

__all__ = ["RunFn", "REGISTRY"]



from __future__ import annotations

import uuid
from typing import Any, Mapping, Optional, Sequence

try:  # pragma: no cover
    from sqlalchemy.inspection import inspect as _sa_inspect  # type: ignore
except Exception:  # pragma: no cover
    _sa_inspect = None  # type: ignore

try:  # pragma: no cover
    from sqlalchemy.sql import ClauseElement as SAClause  # type: ignore
except Exception:  # pragma: no cover
    SAClause = None  # type: ignore


def _ctx_get(ctx: Mapping[str, Any], key: str, default: Any = None) -> Any:
    try:
        return ctx[key]
    except Exception:
        return getattr(ctx, key, default)


def payload(ctx: Mapping[str, Any]) -> Any:
    temp = _ctx_get(ctx, "temp", None)
    raw = _ctx_get(ctx, "payload", None)
    if isinstance(temp, Mapping):
        assembled_values = temp.get("assembled_values")
        if isinstance(assembled_values, Mapping) and isinstance(raw, Mapping):
            merged = dict(raw)
            merged.update(assembled_values)
            return merged

    if isinstance(raw, Mapping):
        return raw
    if isinstance(raw, Sequence) and not isinstance(raw, (str, bytes)):
        return raw
    return {}


def db(ctx: Mapping[str, Any]) -> Any:
    return _ctx_get(ctx, "db")


def request(ctx: Mapping[str, Any]) -> Any:
    return _ctx_get(ctx, "request")


def path_params(ctx: Mapping[str, Any]) -> Mapping[str, Any]:
    params = _ctx_get(ctx, "path_params", None)
    if isinstance(params, Mapping):
        return params

    temp = _ctx_get(ctx, "temp", None)
    if isinstance(temp, Mapping):
        route = temp.get("route")
        if isinstance(route, Mapping):
            route_params = route.get("path_params")
            if isinstance(route_params, Mapping):
                return route_params
    return {}


def _pk_name(model: type) -> str:
    if _sa_inspect is not None:
        try:
            mapper = _sa_inspect(model)
            pk_cols = list(getattr(mapper, "primary_key", []) or [])
            if len(pk_cols) == 1:
                col = pk_cols[0]
                name = getattr(col, "key", None) or getattr(col, "name", None)
                if isinstance(name, str) and name:
                    return name
        except Exception:
            pass

    table = getattr(model, "__table__", None)
    if table is not None:
        try:
            pk = getattr(table, "primary_key", None)
            cols_iter = getattr(pk, "columns", None)
            cols = [col for col in cols_iter] if cols_iter is not None else []
            if len(cols) == 1:
                col = cols[0]
                name = getattr(col, "key", None) or getattr(col, "name", None)
                if isinstance(name, str) and name:
                    return name
        except Exception:
            pass
    return "id"


def _pk_type_info(model: type) -> tuple[Optional[type], Optional[Any]]:
    col = None
    if _sa_inspect is not None:
        try:
            mapper = _sa_inspect(model)
            pk_cols = list(getattr(mapper, "primary_key", []) or [])
            if len(pk_cols) == 1:
                col = pk_cols[0]
        except Exception:
            col = None

    if col is None:
        table = getattr(model, "__table__", None)
        if table is not None:
            try:
                pk = getattr(table, "primary_key", None)
                cols_iter = getattr(pk, "columns", None)
                cols = [item for item in cols_iter] if cols_iter is not None else []
                if len(cols) == 1:
                    col = cols[0]
            except Exception:
                col = None

    if col is None:
        return (None, None)

    try:
        col_type = getattr(col, "type", None)
    except Exception:
        col_type = None

    py_t = None
    try:
        py_t = getattr(col_type, "python_type", None)
    except Exception:
        py_t = None

    return (py_t, col_type)


def _is_uuid_type(py_t: Optional[type], sa_type: Optional[Any]) -> bool:
    if py_t is uuid.UUID:
        return True
    try:
        if getattr(sa_type, "as_uuid", False):
            return True
    except Exception:
        pass
    try:
        type_name = type(sa_type).__name__.lower() if sa_type is not None else ""
        if "uuid" in type_name:
            return True
    except Exception:
        pass
    return False


def _coerce_ident_to_pk_type(model: type, value: Any) -> Any:
    py_t, sa_t = _pk_type_info(model)

    if SAClause is not None and isinstance(value, SAClause):  # pragma: no cover
        return value

    if _is_uuid_type(py_t, sa_t):
        if isinstance(value, uuid.UUID):
            return value
        if isinstance(value, str):
            return uuid.UUID(value)
        if isinstance(value, (bytes, bytearray)) and len(value) == 16:
            return uuid.UUID(bytes=bytes(value))
        return value

    if py_t is int:
        if isinstance(value, int):
            return value
        if isinstance(value, str):
            return int(value)
        try:
            return int(value)  # type: ignore[arg-type]
        except Exception:
            return value

    return value


def _is_clause(value: Any) -> bool:
    return SAClause is not None and isinstance(value, SAClause)  # type: ignore[truthy-bool]


def ident(model: type, ctx: Mapping[str, Any]) -> Any:
    model_payload = payload(ctx)
    model_path = path_params(ctx)
    pk = _pk_name(model)

    candidates: list[tuple[Mapping[str, Any], str]] = [
        (model_path, pk),
        (model_payload, pk),
        (model_path, "id"),
        (model_payload, "id"),
        (model_path, "item_id"),
        (model_payload, "item_id"),
    ]
    if pk != "id":
        candidates.extend([(model_path, f"{pk}_id"), (model_payload, f"{pk}_id")])
    candidates.append((model_payload, "ident"))

    for source, key in candidates:
        try:
            value = source.get(key)
        except Exception:
            value = None
        if value is None or _is_clause(value):
            continue
        try:
            return _coerce_ident_to_pk_type(model, value)
        except Exception as exc:
            raise TypeError(f"Invalid identifier for '{pk}': {value!r}") from exc

    raise TypeError(f"Missing identifier '{pk}' in path or payload")


__all__ = ["db", "ident", "path_params", "payload", "request"]



from __future__ import annotations

import json
from typing import Any

from ... import events as _ev
from ...stages import Ingress
from ...types import Atom, Ctx, IngressCtx
from .._temp import _ensure_temp

ANCHOR = _ev.INGRESS_INPUT_PREPARE


def _run(obj: object | None, ctx: Any) -> None:
    del obj
    temp = _ensure_temp(ctx)
    ingress = temp.setdefault("ingress", {})

    body = getattr(ctx, "body", ingress.get("body"))
    if isinstance(body, (bytes, bytearray, memoryview)):
        raw_bytes = bytes(body)
        ingress["body_peek"] = raw_bytes[:256]
        content_type = str(
            (getattr(ctx, "headers", {}) or {}).get("content-type", "")
        ).lower()
        if "json" in content_type:
            try:
                ingress["body_json"] = json.loads(raw_bytes.decode("utf-8"))
            except Exception:
                pass
    elif body is not None:
        ingress["body_peek"] = str(body)[:256]

    request = getattr(ctx, "request", None)
    if request is not None and body is not None and hasattr(request, "body"):
        request.body = body


class AtomImpl(Atom[Ingress, Ingress]):
    name = "ingress.input_prepare"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Ingress]) -> Ctx[Ingress]:
        _run(obj, ctx)
        return ctx.promote(IngressCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Iterable, List, Tuple

from .phases import (
    EGRESS_FINALIZE_PHASE,
    EGRESS_SHAPE_PHASE,
    END_TX_PHASE,
    HANDLER_PHASE,
    INGRESS_BEGIN_PHASE,
    INGRESS_DISPATCH_PHASE,
    INGRESS_PARSE_PHASE,
    PHASES,
    POST_COMMIT_PHASE,
    POST_HANDLER_PHASE,
    POST_RESPONSE_PHASE,
    PRE_COMMIT_PHASE,
    PRE_HANDLER_PHASE,
    PRE_TX_BEGIN_PHASE,
    START_TX_PHASE,
    PhaseName,
    phase_info,
)
from .stages import (
    Guarded,
    Boot,
    Bound,
    Egressed,
    Emitting,
    Encoded,
    Executing,
    Ingress,
    Operated,
    Resolved,
    Planned,
    Stage,
    stage_ordinal,
)

Phase = PhaseName

INGRESS_BEGIN: Phase = INGRESS_BEGIN_PHASE
INGRESS_PARSE: Phase = INGRESS_PARSE_PHASE
INGRESS_DISPATCH: Phase = INGRESS_DISPATCH_PHASE
PRE_TX_BEGIN: Phase = PRE_TX_BEGIN_PHASE
START_TX: Phase = START_TX_PHASE
PRE_HANDLER: Phase = PRE_HANDLER_PHASE
POST_HANDLER: Phase = POST_HANDLER_PHASE
PRE_COMMIT: Phase = PRE_COMMIT_PHASE
END_TX: Phase = END_TX_PHASE
POST_COMMIT: Phase = POST_COMMIT_PHASE
EGRESS_SHAPE: Phase = EGRESS_SHAPE_PHASE
EGRESS_FINALIZE: Phase = EGRESS_FINALIZE_PHASE
POST_RESPONSE: Phase = POST_RESPONSE_PHASE

INGRESS_CTX_INIT = "ingress.ctx.init"
INGRESS_TRANSPORT_EXTRACT = "ingress.transport.extract"
INGRESS_INPUT_PREPARE = "ingress.input.prepare"

DISPATCH_BINDING_MATCH = "dispatch.binding.match"
DISPATCH_BINDING_PARSE = "dispatch.binding.parse"
DISPATCH_INPUT_NORMALIZE = "dispatch.input.normalize"
DISPATCH_OP_RESOLVE = "dispatch.op.resolve"

DEP_SECURITY = "dep:security"
DEP_EXTRA = "dep:extra"
SYS_TX_BEGIN = "sys.tx.begin"

SCHEMA_COLLECT_IN = "schema:collect_in"
IN_VALIDATE = "in:validate"
RESOLVE_VALUES = "resolve:values"
PRE_FLUSH = "pre:flush"
EMIT_ALIASES_PRE = "emit:aliases:pre_flush"

HANDLER = "HANDLER"
SYS_HANDLER_PERSISTENCE = "sys.handler.persistence"

SYS_TX_COMMIT = "sys.tx.commit"

POST_FLUSH = "post:flush"
EMIT_ALIASES_POST = "emit:aliases:post_refresh"
SCHEMA_COLLECT_OUT = "schema:collect_out"
OUT_BUILD = "out:build"
EMIT_ALIASES_READ = "emit:aliases:readtime"
OUT_DUMP = "out:dump"

EGRESS_RESULT_NORMALIZE = "egress.result.normalize"
EGRESS_OUT_DUMP = "egress.out.dump"
EGRESS_ENVELOPE_APPLY = "egress.envelope.apply"
EGRESS_HEADERS_APPLY = "egress.headers.apply"

EGRESS_HTTP_FINALIZE = "egress.http.finalize"
EGRESS_TO_TRANSPORT_RESPONSE = "egress.to_transport_response"
EGRESS_ASGI_SEND = "egress.asgi.send"

_EVENT_ORDER: Tuple[str, ...] = (
    INGRESS_CTX_INIT,
    INGRESS_TRANSPORT_EXTRACT,
    INGRESS_INPUT_PREPARE,
    DISPATCH_BINDING_MATCH,
    DISPATCH_BINDING_PARSE,
    DISPATCH_INPUT_NORMALIZE,
    DISPATCH_OP_RESOLVE,
    DEP_SECURITY,
    DEP_EXTRA,
    SYS_TX_BEGIN,
    SCHEMA_COLLECT_IN,
    IN_VALIDATE,
    RESOLVE_VALUES,
    PRE_FLUSH,
    EMIT_ALIASES_PRE,
    HANDLER,
    SYS_HANDLER_PERSISTENCE,
    SYS_TX_COMMIT,
    POST_FLUSH,
    EMIT_ALIASES_POST,
    SCHEMA_COLLECT_OUT,
    OUT_BUILD,
    EMIT_ALIASES_READ,
    OUT_DUMP,
    EGRESS_RESULT_NORMALIZE,
    EGRESS_OUT_DUMP,
    EGRESS_ENVELOPE_APPLY,
    EGRESS_HEADERS_APPLY,
    EGRESS_HTTP_FINALIZE,
    EGRESS_TO_TRANSPORT_RESPONSE,
    EGRESS_ASGI_SEND,
)


@dataclass(frozen=True)
class AnchorInfo:
    name: str
    phase: Phase
    ordinal: int
    persist_tied: bool
    stage_in: Stage
    stage_out: Stage
    in_tx: bool
    is_error: bool


_ANCHOR_PHASE: Dict[str, Phase] = {
    INGRESS_CTX_INIT: INGRESS_BEGIN,
    INGRESS_TRANSPORT_EXTRACT: INGRESS_PARSE,
    INGRESS_INPUT_PREPARE: INGRESS_PARSE,
    DISPATCH_BINDING_MATCH: INGRESS_DISPATCH,
    DISPATCH_BINDING_PARSE: INGRESS_DISPATCH,
    DISPATCH_INPUT_NORMALIZE: INGRESS_DISPATCH,
    DISPATCH_OP_RESOLVE: INGRESS_DISPATCH,
    DEP_SECURITY: PRE_TX_BEGIN,
    DEP_EXTRA: PRE_TX_BEGIN,
    SYS_TX_BEGIN: START_TX,
    SCHEMA_COLLECT_IN: PRE_HANDLER,
    IN_VALIDATE: PRE_HANDLER,
    RESOLVE_VALUES: PRE_HANDLER,
    PRE_FLUSH: PRE_HANDLER,
    EMIT_ALIASES_PRE: PRE_HANDLER,
    HANDLER: HANDLER_PHASE,
    SYS_HANDLER_PERSISTENCE: HANDLER_PHASE,
    SYS_TX_COMMIT: END_TX,
    POST_FLUSH: POST_COMMIT,
    EMIT_ALIASES_POST: POST_COMMIT,
    SCHEMA_COLLECT_OUT: POST_COMMIT,
    OUT_BUILD: POST_COMMIT,
    EMIT_ALIASES_READ: POST_COMMIT,
    OUT_DUMP: POST_COMMIT,
    EGRESS_RESULT_NORMALIZE: EGRESS_SHAPE,
    EGRESS_OUT_DUMP: EGRESS_SHAPE,
    EGRESS_ENVELOPE_APPLY: EGRESS_SHAPE,
    EGRESS_HEADERS_APPLY: EGRESS_SHAPE,
    EGRESS_HTTP_FINALIZE: EGRESS_FINALIZE,
    EGRESS_TO_TRANSPORT_RESPONSE: EGRESS_FINALIZE,
    EGRESS_ASGI_SEND: EGRESS_FINALIZE,
}

_ANCHOR_STAGE: Dict[str, Tuple[Stage, Stage]] = {
    INGRESS_CTX_INIT: (Boot, Boot),
    INGRESS_TRANSPORT_EXTRACT: (Ingress, Ingress),
    INGRESS_INPUT_PREPARE: (Ingress, Ingress),
    DISPATCH_BINDING_MATCH: (Ingress, Bound),
    DISPATCH_BINDING_PARSE: (Bound, Bound),
    DISPATCH_INPUT_NORMALIZE: (Bound, Bound),
    DISPATCH_OP_RESOLVE: (Bound, Planned),
    DEP_SECURITY: (Planned, Guarded),
    DEP_EXTRA: (Guarded, Guarded),
    SYS_TX_BEGIN: (Guarded, Executing),
    SCHEMA_COLLECT_IN: (Executing, Executing),
    IN_VALIDATE: (Executing, Executing),
    RESOLVE_VALUES: (Executing, Resolved),
    PRE_FLUSH: (Resolved, Resolved),
    EMIT_ALIASES_PRE: (Resolved, Resolved),
    HANDLER: (Resolved, Resolved),
    SYS_HANDLER_PERSISTENCE: (Resolved, Operated),
    SYS_TX_COMMIT: (Operated, Operated),
    POST_FLUSH: (Operated, Operated),
    EMIT_ALIASES_POST: (Operated, Operated),
    SCHEMA_COLLECT_OUT: (Operated, Operated),
    OUT_BUILD: (Operated, Encoded),
    EMIT_ALIASES_READ: (Encoded, Encoded),
    OUT_DUMP: (Encoded, Encoded),
    EGRESS_RESULT_NORMALIZE: (Encoded, Encoded),
    EGRESS_OUT_DUMP: (Encoded, Encoded),
    EGRESS_ENVELOPE_APPLY: (Encoded, Encoded),
    EGRESS_HEADERS_APPLY: (Encoded, Encoded),
    EGRESS_HTTP_FINALIZE: (Emitting, Emitting),
    EGRESS_TO_TRANSPORT_RESPONSE: (Emitting, Egressed),
    EGRESS_ASGI_SEND: (Egressed, Egressed),
}

_PERSIST_TIED = {
    SYS_TX_BEGIN,
    RESOLVE_VALUES,
    PRE_FLUSH,
    EMIT_ALIASES_PRE,
    SYS_HANDLER_PERSISTENCE,
    SYS_TX_COMMIT,
    POST_FLUSH,
    EMIT_ALIASES_POST,
}


def _validate_anchor_stage_windows() -> None:
    for anchor, (src, dst) in _ANCHOR_STAGE.items():
        if stage_ordinal(src) > stage_ordinal(dst):
            raise ValueError(
                f"Non-monotonic anchor stage window: {anchor!r} ({src.__name__} -> {dst.__name__})"
            )
        pinfo = phase_info(_ANCHOR_PHASE[anchor])
        if stage_ordinal(src) < stage_ordinal(pinfo.stage_in):
            raise ValueError(
                f"Anchor {anchor!r} enters before phase {pinfo.name}: {src.__name__} < {pinfo.stage_in.__name__}"
            )
        if stage_ordinal(dst) > stage_ordinal(pinfo.stage_out):
            raise ValueError(
                f"Anchor {anchor!r} exits after phase {pinfo.name}: {dst.__name__} > {pinfo.stage_out.__name__}"
            )


_validate_anchor_stage_windows()

_ANCHORS: Dict[str, AnchorInfo] = {
    anchor: AnchorInfo(
        name=anchor,
        phase=_ANCHOR_PHASE[anchor],
        ordinal=idx,
        persist_tied=anchor in _PERSIST_TIED,
        stage_in=_ANCHOR_STAGE[anchor][0],
        stage_out=_ANCHOR_STAGE[anchor][1],
        in_tx=phase_info(_ANCHOR_PHASE[anchor]).in_tx,
        is_error=phase_info(_ANCHOR_PHASE[anchor]).is_error,
    )
    for idx, anchor in enumerate(_EVENT_ORDER)
}


def is_valid_event(anchor: str) -> bool:
    return anchor in _ANCHORS


def phase_for_event(anchor: str) -> Phase:
    try:
        return _ANCHORS[anchor].phase
    except KeyError as e:
        raise ValueError(f"Unknown event anchor: {anchor!r}") from e


def stage_in_for_event(anchor: str) -> Stage:
    try:
        return _ANCHORS[anchor].stage_in
    except KeyError as e:
        raise ValueError(f"Unknown event anchor: {anchor!r}") from e


def stage_out_for_event(anchor: str) -> Stage:
    try:
        return _ANCHORS[anchor].stage_out
    except KeyError as e:
        raise ValueError(f"Unknown event anchor: {anchor!r}") from e


def is_persist_tied(anchor: str) -> bool:
    try:
        return _ANCHORS[anchor].persist_tied
    except KeyError as e:
        raise ValueError(f"Unknown event anchor: {anchor!r}") from e


def is_tx_event(anchor: str) -> bool:
    try:
        return _ANCHORS[anchor].in_tx
    except KeyError as e:
        raise ValueError(f"Unknown event anchor: {anchor!r}") from e


def get_anchor_info(anchor: str) -> AnchorInfo:
    try:
        return _ANCHORS[anchor]
    except KeyError as e:
        raise ValueError(f"Unknown event anchor: {anchor!r}") from e


def all_events_ordered() -> List[str]:
    return list(_EVENT_ORDER)


def events_for_phase(phase: Phase) -> List[str]:
    if phase not in PHASES:
        raise ValueError(f"Unknown phase: {phase!r}")
    return [a for a, info in _ANCHORS.items() if info.phase == phase]


def prune_events_for_persist(anchors: Iterable[str], *, persist: bool) -> List[str]:
    out: List[str] = []
    for a in anchors:
        if not is_valid_event(a):
            raise ValueError(f"Unknown event anchor: {a!r}")
        if not persist and _ANCHORS[a].persist_tied:
            continue
        out.append(a)
    out.sort(key=lambda x: _ANCHORS[x].ordinal)
    return out


def order_events(anchors: Iterable[str]) -> List[str]:
    anchors = list(anchors)
    for a in anchors:
        if a not in _ANCHORS:
            raise ValueError(f"Unknown event anchor: {a!r}")
    anchors.sort(key=lambda x: _ANCHORS[x].ordinal)
    return anchors


__all__ = [
    "Phase",
    "PHASES",
    "INGRESS_BEGIN",
    "INGRESS_PARSE",
    "INGRESS_DISPATCH",
    "PRE_TX_BEGIN",
    "START_TX",
    "PRE_HANDLER",
    "HANDLER_PHASE",
    "POST_HANDLER",
    "PRE_COMMIT",
    "END_TX",
    "POST_COMMIT",
    "EGRESS_SHAPE",
    "EGRESS_FINALIZE",
    "POST_RESPONSE",
    "INGRESS_CTX_INIT",
    "INGRESS_TRANSPORT_EXTRACT",
    "INGRESS_INPUT_PREPARE",
    "DISPATCH_BINDING_MATCH",
    "DISPATCH_BINDING_PARSE",
    "DISPATCH_INPUT_NORMALIZE",
    "DISPATCH_OP_RESOLVE",
    "DEP_SECURITY",
    "DEP_EXTRA",
    "SYS_TX_BEGIN",
    "SCHEMA_COLLECT_IN",
    "IN_VALIDATE",
    "RESOLVE_VALUES",
    "PRE_FLUSH",
    "EMIT_ALIASES_PRE",
    "HANDLER",
    "SYS_HANDLER_PERSISTENCE",
    "SYS_TX_COMMIT",
    "POST_FLUSH",
    "EMIT_ALIASES_POST",
    "SCHEMA_COLLECT_OUT",
    "OUT_BUILD",
    "EMIT_ALIASES_READ",
    "OUT_DUMP",
    "EGRESS_RESULT_NORMALIZE",
    "EGRESS_OUT_DUMP",
    "EGRESS_ENVELOPE_APPLY",
    "EGRESS_HEADERS_APPLY",
    "EGRESS_HTTP_FINALIZE",
    "EGRESS_TO_TRANSPORT_RESPONSE",
    "EGRESS_ASGI_SEND",
    "AnchorInfo",
    "is_valid_event",
    "phase_for_event",
    "stage_in_for_event",
    "stage_out_for_event",
    "is_persist_tied",
    "is_tx_event",
    "get_anchor_info",
    "all_events_ordered",
    "events_for_phase",
    "prune_events_for_persist",
    "order_events",
]



from __future__ import annotations

from typing import Any, Callable, Dict, Optional, Tuple

from ... import events as _ev
from . import ctx_init as _ctx_init
from . import input_prepare as _input_prepare
from . import transport_extract as _transport_extract

RunFn = Callable[[Optional[object], Any], Any]

_ORDERED: Tuple[Tuple[str, str, str, RunFn], ...] = (
    ("ingress", "ctx_init", _ctx_init.ANCHOR, _ctx_init.INSTANCE),
    (
        "ingress",
        "transport_extract",
        _transport_extract.ANCHOR,
        _transport_extract.INSTANCE,
    ),
    ("ingress", "input_prepare", _input_prepare.ANCHOR, _input_prepare.INSTANCE),
)

REGISTRY: Dict[Tuple[str, str], Tuple[str, RunFn]] = {
    (domain, subject): (anchor, run) for domain, subject, anchor, run in _ORDERED
}

_EVENT_ORDER = {name: idx for idx, name in enumerate(_ev.all_events_ordered())}
if tuple(REGISTRY):
    _anchors = [anchor for anchor, _run in REGISTRY.values()]
    if _anchors != sorted(_anchors, key=lambda a: _EVENT_ORDER.get(a, 10_000)):
        raise RuntimeError(
            "Ingress atom registry order must follow runtime event order."
        )

__all__ = ["REGISTRY", "RunFn"]



from __future__ import annotations

from typing import Any, Mapping, MutableMapping, Sequence


def _ensure_temp(ctx: Any) -> MutableMapping[str, Any]:
    temp = getattr(ctx, "temp", None)
    if not isinstance(temp, dict):
        temp = {}
        setattr(ctx, "temp", temp)
    return temp


def _record_hydration(
    ctx: Any,
    values: Mapping[str, Any],
    *,
    via_returning: bool = False,
    replace: bool = False,
) -> None:
    """Record DB hydration values and optional RETURNING provenance."""
    temp = _ensure_temp(ctx)
    if via_returning:
        temp["used_returning"] = True
    if not isinstance(values, Mapping):
        return

    hydrated = temp.get("hydrated_values")
    if replace or not isinstance(hydrated, dict):
        temp["hydrated_values"] = dict(values)
        return
    hydrated.update(values)


def _ensure_response_extras(ctx: Any) -> MutableMapping[str, Any]:
    """Return a mutable extras buffer stored in ``ctx.temp['response_extras']``."""
    temp = _ensure_temp(ctx)
    buffer = temp.get("response_extras")
    if not isinstance(buffer, dict):
        buffer = {}
        temp["response_extras"] = buffer
    return buffer


def _add_response_extras(
    ctx: Any,
    extras: Mapping[str, Any],
    *,
    overwrite: bool = False,
) -> Sequence[str]:
    """Merge extras into ``ctx.temp['response_extras']`` and return skipped keys."""
    if not isinstance(extras, Mapping) or not extras:
        return ()

    buffer = _ensure_response_extras(ctx)

    conflicts: list[str] = []
    for key, value in extras.items():
        if (key in buffer) and not overwrite:
            conflicts.append(key)
            continue
        buffer[key] = value
    return tuple(conflicts)


def _response_payload(ctx: Any) -> Any:
    """Return the canonical response payload stored on context."""
    payload = getattr(ctx, "response_payload", None)
    if payload is not None:
        return payload
    temp = getattr(ctx, "temp", None)
    if isinstance(temp, Mapping):
        return temp.get("response_payload")
    return None


def _set_response_payload(ctx: Any, payload: Any) -> None:
    """Set response payload on context and mirror to temp for compatibility."""
    setattr(ctx, "response_payload", payload)
    temp = _ensure_temp(ctx)
    temp["response_payload"] = payload


__all__ = [
    "_ensure_temp",
    "_record_hydration",
    "_ensure_response_extras",
    "_add_response_extras",
    "_response_payload",
    "_set_response_payload",
]



from __future__ import annotations

from ...types import Atom, Ctx, EncodedCtx
from ...stages import Operated, Encoded
from typing import Any, Optional

from ... import events as _ev
from .renderer import ResponseHints
from .renderer import render

ANCHOR = _ev.OUT_DUMP  # "out:dump"


def _is_jsonrpc_request(ctx: Any) -> bool:
    raw = getattr(ctx, "gw_raw", None) or getattr(ctx, "raw", None)
    if getattr(raw, "kind", None) == "jsonrpc":
        return True

    temp = getattr(ctx, "temp", None)
    route = temp.get("route", {}) if isinstance(temp, dict) else {}
    rpc_env = route.get("rpc_envelope") if isinstance(route, dict) else None
    if isinstance(rpc_env, dict) and rpc_env.get("jsonrpc") == "2.0":
        return True

    path = getattr(raw, "path", None)
    app = getattr(ctx, "app", None)
    prefix = getattr(app, "jsonrpc_prefix", None)
    if isinstance(path, str) and isinstance(prefix, str):
        return (path.rstrip("/") or "/") == (prefix.rstrip("/") or "/")

    return False


def _run(obj: Optional[object], ctx: Any) -> Any:
    """response:render@out:dump

    Render ``ctx.response.result`` into a concrete Response object.
    """
    resp_ns = getattr(ctx, "response", None)
    req = getattr(ctx, "request", None)
    if resp_ns is None or req is None:
        return None
    result = getattr(resp_ns, "result", None)
    if result is None:
        return None
    hints = getattr(resp_ns, "hints", None)
    if hints is None:
        hints = ResponseHints(status_code=int(getattr(ctx, "status_code", 200) or 200))
    default_media = getattr(resp_ns, "default_media", "application/json")
    envelope_default = getattr(resp_ns, "envelope_default", False)
    temp = getattr(ctx, "temp", None)
    route_ns = temp.get("route") if isinstance(temp, dict) else None
    rpc_env = route_ns.get("rpc_envelope") if isinstance(route_ns, dict) else None
    if getattr(getattr(ctx, "gw_raw", None), "kind", None) == "jsonrpc" or (
        isinstance(rpc_env, dict) and rpc_env.get("jsonrpc") == "2.0"
    ):
        return None
    resp = render(
        req,
        result,
        hints=hints,
        default_media=default_media,
        envelope_default=envelope_default,
    )
    resp_ns.result = resp
    temp = getattr(ctx, "temp", None)
    if isinstance(temp, dict):
        egress = temp.setdefault("egress", {})
        if isinstance(egress, dict):
            if _is_jsonrpc_request(ctx) and isinstance(
                egress.get("transport_response"), dict
            ):
                return resp
            egress["transport_response"] = {
                "status_code": (
                    204
                    if int(resp.status_code) == 200 and not (resp.body or b"")
                    else int(resp.status_code)
                ),
                "headers": {
                    k.decode("latin-1"): v.decode("latin-1")
                    for k, v in getattr(resp, "raw_headers", ())
                },
                "body": resp.body or None,
            }
    return resp


class AtomImpl(Atom[Operated, Encoded]):
    name = "response.render"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Operated]) -> Ctx[Encoded]:
        _run(obj, ctx)
        return ctx.promote(EncodedCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



from __future__ import annotations

from typing import Any, Mapping

import tigrbl_ops_oltp as _core
from tigrbl_ops_oltp.crud.helpers.model import _coerce_pk_value

from ... import events as _ev
from ...stages import Operated, Resolved
from ...types import Atom, Ctx, OperatedCtx
from . import _oltp_context as _ctx

ANCHOR = _ev.SYS_HANDLER_PERSISTENCE


async def _run(obj: object | None, ctx: Any) -> None:
    model = obj if isinstance(obj, type) else getattr(ctx, "model", None)
    if not isinstance(model, type):
        raise TypeError("handler_bulk_delete requires a model type")
    payload = _ctx.payload(ctx)
    raw_ids = payload
    if isinstance(payload, Mapping):
        raw_ids = payload.get("ids")
        if raw_ids is None and payload:
            first_value = next(iter(payload.values()))
            raw_ids = first_value if isinstance(first_value, list) else []
    ids: list[Any] = []
    if isinstance(raw_ids, list):
        for ident in raw_ids:
            try:
                ids.append(_coerce_pk_value(model, ident))
            except Exception:
                continue
    setattr(ctx, "result", await _core.bulk_delete(model, ids, db=_ctx.db(ctx)))


class AtomImpl(Atom[Resolved, Operated]):
    name = "sys.handler_bulk_delete"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Resolved]) -> Ctx[Operated]:
        await _run(obj, ctx)
        return ctx.promote(OperatedCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



from __future__ import annotations

from collections.abc import Mapping
from typing import Any


def _ensure_temp(ctx: Any) -> dict[str, Any]:
    tmp = getattr(ctx, "temp", None)
    if not isinstance(tmp, dict):
        tmp = {}
        setattr(ctx, "temp", tmp)
    return tmp


def _ensure_ov(ctx: Any):
    ov = getattr(ctx, "opview", None)
    if ov is None:
        raise RuntimeError("ctx_missing:opview")
    return ov


def _normalize_schema_from_specs(ctx: Any) -> None:
    specs = getattr(ctx, "specs", None)
    op = getattr(ctx, "op", None)
    if not isinstance(specs, Mapping) or not isinstance(op, str) or not op:
        raise RuntimeError("ctx_missing:opview")

    in_fields: list[str] = []
    out_fields: list[str] = []
    by_field_in: dict[str, dict[str, Any]] = {}
    by_field_out: dict[str, dict[str, Any]] = {}

    for field_name, spec in specs.items():
        if not isinstance(field_name, str):
            continue
        io = getattr(spec, "io", None)
        fs = getattr(spec, "field", None)
        storage = getattr(spec, "storage", None)

        in_verbs = set(getattr(io, "in_verbs", ()) or ())
        out_verbs = set(getattr(io, "out_verbs", ()) or ())

        if op in in_verbs:
            in_fields.append(field_name)
            in_meta: dict[str, Any] = {"in_enabled": True}
            if storage is None:
                in_meta["virtual"] = True

            default_factory = getattr(spec, "default_factory", None)
            if callable(default_factory):
                in_meta["default_factory"] = default_factory

            alias_in = getattr(io, "alias_in", None)
            if alias_in:
                in_meta["alias_in"] = alias_in

            header_in = getattr(io, "header_in", None)
            if header_in:
                in_meta["header_in"] = header_in
                in_meta["header_required_in"] = bool(
                    getattr(io, "header_required_in", False)
                )

            in_meta["required"] = bool(fs and op in getattr(fs, "required_in", ()))
            in_meta["nullable"] = (
                True if storage is None else bool(getattr(storage, "nullable", True))
            )
            by_field_in[field_name] = in_meta

        if op in out_verbs:
            out_fields.append(field_name)
            out_meta: dict[str, Any] = {}

            alias_out = getattr(io, "alias_out", None)
            if alias_out:
                out_meta["alias_out"] = alias_out
            if storage is None:
                out_meta["virtual"] = True

            py_type = getattr(getattr(fs, "py_type", None), "__name__", None)
            if py_type:
                out_meta["py_type"] = py_type
            by_field_out[field_name] = out_meta

    in_fields_sorted = tuple(sorted(in_fields))
    out_fields_sorted = tuple(sorted(out_fields))

    setattr(
        ctx,
        "schema_in",
        {
            "fields": in_fields_sorted,
            "by_field": {f: by_field_in.get(f, {}) for f in in_fields_sorted},
            "required": tuple(
                f for f in in_fields_sorted if by_field_in.get(f, {}).get("required")
            ),
        },
    )
    setattr(
        ctx,
        "schema_out",
        {
            "fields": out_fields_sorted,
            "by_field": {f: by_field_out.get(f, {}) for f in out_fields_sorted},
            "expose": out_fields_sorted,
        },
    )


def _ensure_schema_in(ctx: Any) -> Mapping[str, Any]:
    temp = _ensure_temp(ctx)
    cached = temp.get("schema_in")
    if isinstance(cached, Mapping):
        return cached

    schema_in = getattr(ctx, "schema_in", None)
    if isinstance(schema_in, Mapping):
        temp["schema_in"] = schema_in
        return schema_in

    try:
        ov = _ensure_ov(ctx)
        bf = ov.schema_in.by_field
        req = tuple(n for n, e in bf.items() if e.get("required"))
        temp["schema_in"] = {
            "fields": ov.schema_in.fields,
            "by_field": bf,
            "required": req,
        }
    except RuntimeError as exc:
        if str(exc) != "ctx_missing:opview":
            raise
        _normalize_schema_from_specs(ctx)
        temp["schema_in"] = getattr(ctx, "schema_in")
    return temp["schema_in"]


def _ensure_schema_out(ctx: Any) -> Mapping[str, Any]:
    temp = _ensure_temp(ctx)
    cached = temp.get("schema_out")
    if isinstance(cached, Mapping):
        return cached

    schema_out = getattr(ctx, "schema_out", None)
    if isinstance(schema_out, Mapping):
        temp["schema_out"] = schema_out
        return schema_out

    try:
        ov = _ensure_ov(ctx)
        temp["schema_out"] = {
            "fields": ov.schema_out.fields,
            "by_field": ov.schema_out.by_field,
            "expose": ov.schema_out.expose,
        }
    except RuntimeError as exc:
        if str(exc) != "ctx_missing:opview":
            raise
        _normalize_schema_from_specs(ctx)
        temp["schema_out"] = getattr(ctx, "schema_out")
    return temp["schema_out"]



from __future__ import annotations

from ...types import Atom, Ctx, BootCtx
from ...stages import Boot

from time import perf_counter
from typing import Any

from ... import events as _ev
from .._temp import _ensure_temp

ANCHOR = _ev.INGRESS_CTX_INIT


def _run(obj: object | None, ctx: Any) -> None:
    del obj
    temp = _ensure_temp(ctx)
    ingress = temp.setdefault("ingress", {})
    ingress.setdefault("ctx_initialized", True)
    ingress.setdefault("started_at", perf_counter())


class AtomImpl(Atom[Boot, Boot]):
    name = "ingress.ctx_init"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Boot]) -> Ctx[Boot]:
        _run(obj, ctx)
        return ctx.promote(BootCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



from __future__ import annotations

from typing import Any, Callable, Dict, Optional, Tuple

from . import commit_tx as _commit_tx
from . import handler_bulk_create as _handler_bulk_create
from . import handler_bulk_delete as _handler_bulk_delete
from . import handler_bulk_merge as _handler_bulk_merge
from . import handler_bulk_replace as _handler_bulk_replace
from . import handler_bulk_update as _handler_bulk_update
from . import handler_clear as _handler_clear
from . import handler_create as _handler_create
from . import handler_delete as _handler_delete
from . import handler_list as _handler_list
from . import handler_merge as _handler_merge
from . import handler_noop as _handler_noop
from . import handler_persistence as _handler_persistence
from . import handler_read as _handler_read
from . import handler_replace as _handler_replace
from . import handler_update as _handler_update
from . import start_tx as _start_tx

RunFn = Callable[[Optional[object], Any], Any]

_ORDERED: Tuple[Tuple[str, str, str, RunFn], ...] = (
    ("sys", "start_tx", _start_tx.ANCHOR, _start_tx.INSTANCE),
    (
        "sys",
        "handler_persistence",
        _handler_persistence.ANCHOR,
        _handler_persistence.INSTANCE,
    ),
    ("sys", "handler_create", _handler_create.ANCHOR, _handler_create.INSTANCE),
    ("sys", "handler_read", _handler_read.ANCHOR, _handler_read.INSTANCE),
    ("sys", "handler_update", _handler_update.ANCHOR, _handler_update.INSTANCE),
    ("sys", "handler_replace", _handler_replace.ANCHOR, _handler_replace.INSTANCE),
    ("sys", "handler_merge", _handler_merge.ANCHOR, _handler_merge.INSTANCE),
    ("sys", "handler_noop", _handler_noop.ANCHOR, _handler_noop.INSTANCE),
    ("sys", "handler_delete", _handler_delete.ANCHOR, _handler_delete.INSTANCE),
    ("sys", "handler_list", _handler_list.ANCHOR, _handler_list.INSTANCE),
    ("sys", "handler_clear", _handler_clear.ANCHOR, _handler_clear.INSTANCE),
    (
        "sys",
        "handler_bulk_create",
        _handler_bulk_create.ANCHOR,
        _handler_bulk_create.INSTANCE,
    ),
    (
        "sys",
        "handler_bulk_update",
        _handler_bulk_update.ANCHOR,
        _handler_bulk_update.INSTANCE,
    ),
    (
        "sys",
        "handler_bulk_replace",
        _handler_bulk_replace.ANCHOR,
        _handler_bulk_replace.INSTANCE,
    ),
    (
        "sys",
        "handler_bulk_merge",
        _handler_bulk_merge.ANCHOR,
        _handler_bulk_merge.INSTANCE,
    ),
    (
        "sys",
        "handler_bulk_delete",
        _handler_bulk_delete.ANCHOR,
        _handler_bulk_delete.INSTANCE,
    ),
    ("sys", "commit_tx", _commit_tx.ANCHOR, _commit_tx.INSTANCE),
)

REGISTRY: Dict[Tuple[str, str], Tuple[str, RunFn]] = {
    (domain, subject): (anchor, run) for domain, subject, anchor, run in _ORDERED
}

__all__ = ["REGISTRY", "RunFn"]



from __future__ import annotations

from .types import (
    EGRESS_PHASES,
    HANDLER_PHASES,
    INGRESS_PHASES,
    PHASE_SEQUENCE,
    HookPhase,
    HookPhases,
    HookPredicate,
    StepFn,
    VALID_HOOK_PHASES,
)

__all__ = [
    "PHASE_SEQUENCE",
    "INGRESS_PHASES",
    "HANDLER_PHASES",
    "EGRESS_PHASES",
    "HookPhase",
    "HookPhases",
    "VALID_HOOK_PHASES",
    "StepFn",
    "HookPredicate",
]



from __future__ import annotations

from typing import Any

from ... import events as _ev
from ...stages import Operated, Resolved
from ...types import Atom, Ctx, OperatedCtx

ANCHOR = _ev.SYS_HANDLER_PERSISTENCE
TARGET = "noop"


async def _run(obj: object | None, ctx: Any) -> None:
    del obj
    setattr(
        ctx,
        "result",
        {
            "ok": True,
            "noop": True,
            "alias": getattr(ctx, "op", None),
            "target": TARGET,
        },
    )


class AtomImpl(Atom[Resolved, Operated]):
    name = "handler.noop"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Resolved]) -> Ctx[Operated]:
        await _run(obj, ctx)
        return ctx.promote(OperatedCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "TARGET", "INSTANCE"]



from __future__ import annotations

from typing import Any

from ...types import Atom, Ctx, ExecutingCtx
from ...stages import Guarded, Executing

from ... import events as _ev
from .._temp import _ensure_temp
from ._db import _is_async_db, _resolve_db_handle

ANCHOR = _ev.SYS_TX_BEGIN


async def _run(obj: object | None, ctx: Any) -> None:
    del obj
    temp = _ensure_temp(ctx)
    db = _resolve_db_handle(ctx)
    temp["__sys_tx_open__"] = False

    if db is None:
        return

    # Sync SQLAlchemy sessions auto-open transactions on first write.
    # Avoid an explicit begin() call in hot paths to reduce per-request
    # transaction setup overhead.
    if not _is_async_db(db):
        return

    begin = getattr(db, "begin", None)
    if callable(begin):
        rv = begin()
        if hasattr(rv, "__await__"):
            await rv
        temp["__sys_tx_open__"] = True


class AtomImpl(Atom[Guarded, Executing]):
    name = "sys.start_tx"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Guarded]) -> Ctx[Executing]:
        await _run(obj, ctx)
        return ctx.promote(ExecutingCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



# pkgs/standards/tigrbl_atoms/tigrbl/atoms/resolve/assemble.py
from __future__ import annotations

from ...types import Atom, Ctx, ResolvedCtx
from ...stages import Executing, Resolved

from typing import Any, Mapping, Optional, Dict, Tuple
import logging

from ... import events as _ev
from ..._ctx import _ctx_view
from ..._opview_helpers import _ensure_schema_in, _ensure_temp

# Runs in HANDLER phase, before pre:flush and any storage transforms.
ANCHOR = _ev.RESOLVE_VALUES  # "resolve:values"

logger = logging.getLogger("uvicorn")


def _run(obj: Optional[object], ctx: Any) -> None:
    """
    resolve:assemble@resolve:values

    Purpose
    -------
    Build a normalized dict of inbound values to apply on the model (assembled_values).
    - Prefer client-provided, validated values (from wire:build_in/validate).
    - For fields that are **ABSENT** (not present in inbound), apply ColumnSpec.default_factory(ctx)
      for simple server-side defaults (non-paired path).
    - Virtual columns (storage=None) are never persisted; if present in inbound, stash them
      under temp["virtual_in"] for downstream wire/out logic.

    Inputs (conventions)
    --------------------
    - ctx.temp["in_values"] OR ctx.in_data / ctx.payload / ctx.data / ctx.body :
        dict-like or Pydantic model; used as inbound source
    - ctx.persist : bool  (writes only; non-persist ops typically prune this anchor)
    - ctx.op      : str   (verb name; optional, only for diagnostics)

    Effects
    -------
    - ctx.temp["assembled_values"] : dict[field -> value] (only persisted fields)
    - ctx.temp["virtual_in"]       : dict[field -> value] (for storage=None)
    - ctx.temp["absent_fields"]    : tuple[str, ...]      (those not in inbound)
    - ctx.temp["used_default_factory"] : tuple[str, ...]  (fields defaulted here)
    """
    # Non-persisting ops should have pruned this anchor; retain guard for safety.
    if getattr(ctx, "persist", True) is False:
        logger.debug("Skipping resolve:assemble; ctx.persist is False")
        return

    logger.debug("Running resolve:assemble")
    schema_in = _ensure_schema_in(ctx)
    inbound = _coerce_inbound(getattr(ctx, "temp", {}).get("in_values", None), ctx)

    temp = _ensure_temp(ctx)
    assembled: Dict[str, Any] = {}
    virtual_in: Dict[str, Any] = {}
    absent: list[str] = []
    used_default: list[str] = []

    # Iterate fields in a stable order
    for field in sorted(schema_in["fields"]):
        meta = schema_in["by_field"].get(field, {})
        in_enabled = meta.get("in_enabled", True)
        is_virtual = meta.get("virtual", False)

        present, value = _try_read_inbound(inbound, field)
        if present:
            if is_virtual:
                virtual_in[field] = value
                logger.debug("Captured virtual inbound %s=%s", field, value)
            elif in_enabled:
                assembled[field] = value
                logger.debug("Assembled inbound %s=%s", field, value)
            continue

        absent.append(field)
        logger.debug("Field %s absent from inbound", field)

        default_fn = meta.get("default_factory")
        if callable(default_fn) and in_enabled and not is_virtual:
            try:
                default_val = default_fn(_ctx_view(ctx))
                assembled[field] = default_val
                used_default.append(field)
                logger.debug("Applied default for field %s", field)
            except Exception:
                logger.debug("Default factory failed for field %s", field)

    # Stash results on ctx.temp
    temp["assembled_values"] = assembled
    temp["virtual_in"] = virtual_in
    temp["absent_fields"] = tuple(absent)
    temp["used_default_factory"] = tuple(used_default)
    logger.debug(
        "Assembled values: %s, virtual_in: %s, absent: %s, defaults: %s",
        assembled,
        virtual_in,
        absent,
        used_default,
    )


# ──────────────────────────────────────────────────────────────────────────────
# Internals
# ──────────────────────────────────────────────────────────────────────────────


def _coerce_inbound(candidate: Any, ctx: Any) -> Mapping[str, Any]:
    """
    Return a dict-like for inbound values.
    Priority: ctx.temp["in_values"] → ctx.in_data → ctx.payload → ctx.data → ctx.body.
    Accepts Pydantic v1/v2 models (model_dump()/dict()).
    """
    for name in ("in_values",):
        if isinstance(candidate, Mapping):
            return candidate
    # fallbacks on ctx
    for attr in ("in_data", "payload", "data", "body"):
        val = getattr(ctx, attr, None)
        if isinstance(val, Mapping):
            return val
        # Pydantic v2
        if hasattr(val, "model_dump") and callable(getattr(val, "model_dump")):
            try:
                return dict(val.model_dump())
            except Exception:
                pass
        # Pydantic v1
        if hasattr(val, "dict") and callable(getattr(val, "dict")):
            try:
                return dict(val.dict())
            except Exception:
                pass
    # default empty
    return {}


def _try_read_inbound(inbound: Mapping[str, Any], field: str) -> Tuple[bool, Any]:
    """
    Distinguish ABSENT vs present(None).
    """
    if field in inbound:
        return True, inbound.get(field, None)
    # Be tolerant to alias-style inputs (if present)
    for alt in (field.lower(), field.upper()):
        if alt in inbound:
            return True, inbound.get(alt)
    return False, None


class AtomImpl(Atom[Executing, Resolved]):
    name = "resolve.assemble"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Executing]) -> Ctx[Resolved]:
        _run(obj, ctx)
        return ctx.promote(ResolvedCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



# pkgs/standards/tigrbl_atoms/tigrbl/atoms/__init__.py
from __future__ import annotations

from typing import Any, Callable, Dict, Optional, Tuple
import logging

from .. import events as _ev

# Domain registries
from .emit import REGISTRY as _EMIT
from .out import REGISTRY as _OUT
from .refresh import REGISTRY as _REFRESH
from .resolve import REGISTRY as _RESOLVE
from .schema import REGISTRY as _SCHEMA
from .storage import REGISTRY as _STORAGE
from .wire import REGISTRY as _WIRE
from .response import REGISTRY as _RESPONSE
from .dep import REGISTRY as _DEP
from .egress import REGISTRY as _EGRESS
from .ingress import REGISTRY as _INGRESS
from .dispatch import REGISTRY as _DISPATCH
from .sys import REGISTRY as _SYS

# Runner signature: (obj|None, ctx) -> None
RunFn = Callable[[Optional[object], Any], Any]

#: Global registry consumed by the kernel plan:
#:   { (domain, subject): (anchor, runner) }
REGISTRY: Dict[Tuple[str, str], Tuple[str, RunFn]] = {}

logger = logging.getLogger("uvicorn")


def _add_bulk(source: Dict[Tuple[str, str], Tuple[str, RunFn]]) -> None:
    for key, val in source.items():
        if key in REGISTRY:
            logger.error("Duplicate atom registration attempted: %s", key)
            raise RuntimeError(f"Duplicate atom registration: {key!r}")
        anchor, fn = val
        if not _ev.is_valid_event(anchor):
            logger.error("Atom %s declares unknown anchor %s", key, anchor)
            raise ValueError(f"Atom {key!r} declares unknown anchor {anchor!r}")
        REGISTRY[key] = (anchor, fn)
        logger.debug("Registered atom %s -> %s", key, anchor)


# Aggregate all domains
_add_bulk(_EMIT)
_add_bulk(_OUT)
_add_bulk(_REFRESH)
_add_bulk(_RESOLVE)
_add_bulk(_SCHEMA)
_add_bulk(_STORAGE)
_add_bulk(_WIRE)
_add_bulk(_RESPONSE)
_add_bulk(_DEP)
_add_bulk(_INGRESS)
_add_bulk(_DISPATCH)
_add_bulk(_EGRESS)
_add_bulk(_SYS)

logger.info("Loaded %d runtime atoms", len(REGISTRY))

# ── Back-compat subject aliases (optional) ────────────────────────────────────
# Allow "wire:validate" as an alias of "wire:validate_in".
if ("wire", "validate_in") in REGISTRY and ("wire", "validate") not in REGISTRY:
    REGISTRY[("wire", "validate")] = REGISTRY[("wire", "validate_in")]

# ── Public helpers ────────────────────────────────────────────────────────────


def domains() -> Tuple[str, ...]:
    """Return all domains present in the registry."""
    out = tuple(sorted({d for (d, _) in REGISTRY.keys()}))
    logger.debug("Listing domains: %s", out)
    return out


def subjects(domain: str) -> Tuple[str, ...]:
    """Return subjects available for a given domain."""
    out = tuple(sorted(s for (d, s) in REGISTRY.keys() if d == domain))
    logger.debug("Listing subjects for %s: %s", domain, out)
    return out


def get(domain: str, subject: str) -> Tuple[str, RunFn]:
    """Return (anchor, runner) for a given (domain, subject)."""
    key = (domain, subject)
    if key not in REGISTRY:
        logger.error("Unknown atom requested: %s:%s", domain, subject)
        raise KeyError(f"Unknown atom: {domain}:{subject}")
    val = REGISTRY[key]
    logger.debug("Retrieved atom %s:%s -> %s", domain, subject, val[0])
    return val


def all_items() -> Tuple[Tuple[Tuple[str, str], Tuple[str, RunFn]], ...]:
    """Return the registry items as a sorted tuple for deterministic iteration."""
    items = tuple(sorted(REGISTRY.items(), key=lambda kv: (kv[0][0], kv[0][1])))
    logger.debug("Listing all registry items (%d)", len(items))
    return items


__all__ = [
    "RunFn",
    "REGISTRY",
    "domains",
    "subjects",
    "get",
    "all_items",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import MISSING, dataclass, field, fields, is_dataclass
from enum import Enum
from typing import Any, Awaitable, Callable, Final, Tuple, cast, final

from collections.abc import Mapping

from typing_extensions import Generic, TypeAlias, TypeVar

from .stages import (
    Boot,
    Failed,
)

from tigrbl_typing.protocols import (
    DependencyLike,
    ResponseLike,
    is_dependency_like,
    is_response_like,
)


S = TypeVar("S")
T = TypeVar("T")
U = TypeVar("U")
E = TypeVar("E", default=Exception)


PHASE_SEQUENCE: Final[tuple[str, ...]] = (
    "INGRESS_BEGIN",
    "INGRESS_PARSE",
    "INGRESS_DISPATCH",
    "PRE_TX_BEGIN",
    "START_TX",
    "PRE_HANDLER",
    "HANDLER",
    "POST_HANDLER",
    "PRE_COMMIT",
    "END_TX",
    "POST_COMMIT",
    "POST_RESPONSE",
    "EGRESS_SHAPE",
    "EGRESS_FINALIZE",
)

INGRESS_PHASES: Final[tuple[str, ...]] = PHASE_SEQUENCE[:4]
HANDLER_PHASES: Final[tuple[str, ...]] = PHASE_SEQUENCE[5:8]
EGRESS_PHASES: Final[tuple[str, ...]] = PHASE_SEQUENCE[-3:]


@dataclass(slots=True)
class BaseCtx(Generic[S, E]):
    env: object | None = None
    bag: dict[str, object] = field(default_factory=dict)
    temp: dict[str, object] = field(default_factory=dict)
    error: E | None = None
    current_phase: str | None = None
    error_phase: str | None = None

    def __getattribute__(self, name: str) -> Any:
        """
        Keep core context methods callable even if runtime data shadows names.

        Some adapters may attach arbitrary attributes to ctx instances. If one of
        those attributes is named ``promote`` and set to ``None`` (or any
        non-callable), ``ctx.promote(...)`` would fail with ``NoneType is not
        callable``. Always resolving ``promote`` from ``BaseCtx`` preserves the
        invariant that every ctx remains promotable.
        """
        if name == "promote":
            return BaseCtx.promote.__get__(self, type(self))
        return object.__getattribute__(self, name)

    def promote(self, cls: type[U], /, **updates: object) -> U:
        if not is_dataclass(cls):
            raise TypeError(f"promote target must be a dataclass type, got {cls!r}")

        data: dict[str, object] = {}
        missing_required: list[str] = []

        for f in fields(cls):
            if f.name in updates:
                continue
            if hasattr(self, f.name):
                data[f.name] = getattr(self, f.name)
                continue
            if f.default is MISSING and f.default_factory is MISSING:
                missing_required.append(f.name)

        if missing_required:
            raise TypeError(
                f"cannot promote {type(self).__name__} -> {cls.__name__}; "
                f"missing required fields: {', '.join(missing_required)}"
            )

        data.update(updates)
        return cls(**data)

    def put_temp(self, key: str, value: object) -> None:
        self.temp[key] = value

    def require_temp(self, key: str) -> object:
        try:
            return self.temp[key]
        except KeyError as e:
            raise KeyError(f"missing temp field: {key!r}") from e

    def safe_view(
        self,
        *,
        include_temp: bool = False,
        temp_keys: tuple[str, ...] | None = None,
    ) -> Mapping[str, object]:
        """
        Return a compact read-only mapping intended for user callables.

        By default, temp is excluded to avoid exposing runtime internals.
        """
        base: dict[str, object] = {
            "op": getattr(self, "op", None),
            "persist": getattr(self, "persist", True),
            "model": getattr(self, "model", None),
            "specs": getattr(self, "specs", None),
            "user": getattr(self, "user", None),
            "tenant": getattr(self, "tenant", None),
            "now": getattr(self, "now", None),
        }
        if include_temp:
            allowed = set(temp_keys or ("assembled_values", "virtual_in"))
            exposed: dict[str, object] = {}
            for key in allowed:
                if key in self.temp:
                    exposed[key] = self.temp[key]
            base["temp"] = MappingProxy(exposed)
        return MappingProxy(base)


class MappingProxy(Mapping[str, object]):
    __slots__ = ("_d",)

    def __init__(self, data: Mapping[str, object]):
        self._d = dict(data)

    def __getitem__(self, key: str) -> object:
        return self._d[key]

    def __iter__(self):
        return iter(self._d)

    def __len__(self) -> int:
        return len(self._d)

    def get(self, key: str, default: object = None) -> object:
        return self._d.get(key, default)


Ctx: TypeAlias = BaseCtx[S, E]


class HookPhase(str, Enum):
    PRE_TX_BEGIN = "PRE_TX_BEGIN"
    START_TX = "START_TX"
    PRE_HANDLER = "PRE_HANDLER"
    HANDLER = "HANDLER"
    POST_HANDLER = "POST_HANDLER"
    PRE_COMMIT = "PRE_COMMIT"
    END_TX = "END_TX"
    POST_COMMIT = "POST_COMMIT"
    POST_RESPONSE = "POST_RESPONSE"
    ON_ERROR = "ON_ERROR"
    ON_PRE_TX_BEGIN_ERROR = "ON_PRE_TX_BEGIN_ERROR"
    ON_START_TX_ERROR = "ON_START_TX_ERROR"
    ON_PRE_HANDLER_ERROR = "ON_PRE_HANDLER_ERROR"
    ON_HANDLER_ERROR = "ON_HANDLER_ERROR"
    ON_POST_HANDLER_ERROR = "ON_POST_HANDLER_ERROR"
    ON_PRE_COMMIT_ERROR = "ON_PRE_COMMIT_ERROR"
    ON_END_TX_ERROR = "ON_END_TX_ERROR"
    ON_POST_COMMIT_ERROR = "ON_POST_COMMIT_ERROR"
    ON_POST_RESPONSE_ERROR = "ON_POST_RESPONSE_ERROR"
    ON_ROLLBACK = "ON_ROLLBACK"


HookPhases: Tuple[HookPhase, ...] = tuple(HookPhase)
VALID_HOOK_PHASES: set[str] = {hook_phase.value for hook_phase in HookPhases}
StepFn = Callable[[Ctx[Any, Exception]], Awaitable[Any] | Any]
HookPredicate = Callable[[Ctx[Any, Exception]], bool]


def promote(ctx: Ctx[S, E], cls: type[U], /, **updates: object) -> U:
    return ctx.promote(cls, **updates)


def has_error(ctx: BaseCtx[S, E]) -> bool:
    return ctx.error is not None


@dataclass(slots=True)
class BootCtx(BaseCtx[Boot, E], Generic[E]):
    pass


@dataclass(slots=True)
class IngressCtx(BootCtx[E], Generic[E]):
    raw: object | None = None
    request: object | None = None
    method: str = ""
    path: str = ""
    headers: dict[str, object] = field(default_factory=dict)
    query: dict[str, object] = field(default_factory=dict)
    body: object | None = None


@dataclass(slots=True)
class RoutedCtx(IngressCtx[E], Generic[E]):
    protocol: str = ""
    path_params: dict[str, object] = field(default_factory=dict)


@dataclass(slots=True)
class BoundCtx(RoutedCtx[E], Generic[E]):
    binding: object | None = None
    model: object | None = None
    op: str = ""


@dataclass(slots=True)
class PlannedCtx(BoundCtx[E], Generic[E]):
    opmeta_index: object | None = None
    payload: object | None = None
    plan: object | None = None
    opmeta: object | None = None
    opview: object | None = None


@dataclass(slots=True)
class GuardedCtx(PlannedCtx[E], Generic[E]):
    authz: object | None = None


@dataclass(slots=True)
class ExecutingCtx(GuardedCtx[E], Generic[E]):
    pass


@dataclass(slots=True)
class ResolvedCtx(ExecutingCtx[E], Generic[E]):
    schema_in: object | None = None
    in_values: dict[str, object] = field(default_factory=dict)


@dataclass(slots=True)
class OperatedCtx(ResolvedCtx[E], Generic[E]):
    result: object | None = None


@dataclass(slots=True)
class EncodedCtx(OperatedCtx[E], Generic[E]):
    schema_out: object | None = None
    response_payload: object | None = None
    response_headers: dict[str, object] = field(default_factory=dict)
    status_code: int = 200


@dataclass(slots=True)
class EmittingCtx(EncodedCtx[E], Generic[E]):
    transport_response: object | None = None


@dataclass(slots=True)
class EgressedCtx(EmittingCtx[E], Generic[E]):
    pass


@dataclass(slots=True)
class FailedCtx(BaseCtx[Failed, E], Generic[E]):
    error: E | None = None


AtomResult: TypeAlias = Ctx[T, E] | FailedCtx[E]


def fail(ctx: Ctx[S, E], error: E, /, **updates: object) -> FailedCtx[E]:
    return cast(
        FailedCtx[E],
        ctx.promote(
            FailedCtx,
            error=error,
            **updates,
        ),
    )


class AtomFailure(Exception):
    """
    Optional internal exception type for atoms that want to signal
    a mapped domain/runtime failure via raise instead of returning fail(...).
    """

    def __init__(self, error: object) -> None:
        super().__init__(str(error))
        self.error = error


class Atom(ABC, Generic[S, T, E]):
    name: str = "atom"
    anchor: str = ""

    @abstractmethod
    async def __call__(
        self,
        obj: object | None,
        ctx: Ctx[S, E],
    ) -> AtomResult[T, E]:
        raise NotImplementedError


class StandardAtom(Atom[S, T, E], ABC):
    """
    Template-method base:
      - handles pre-failed input
      - handles success promotion
      - handles fail normalization
      - optionally maps raised exceptions into E
    """

    target_ctx: type[BaseCtx[T, E]]

    @final
    async def __call__(
        self,
        obj: object | None,
        ctx: Ctx[S, E],
    ) -> AtomResult[T, E]:
        if has_error(ctx):
            assert ctx.error is not None
            return fail(ctx, ctx.error)

        try:
            result = await self._run(obj, ctx)
        except AtomFailure as ex:
            return fail(ctx, cast(E, ex.error))
        except Exception as ex:
            mapped = self._map_exception(ex)
            if mapped is None:
                raise
            return fail(ctx, mapped)

        if isinstance(result, FailedCtx):
            return result

        return promote(ctx, self.target_ctx, **result)

    @abstractmethod
    async def _run(
        self,
        obj: object | None,
        ctx: Ctx[S, E],
    ) -> dict[str, object] | FailedCtx[E]:
        raise NotImplementedError

    def _map_exception(self, ex: Exception) -> E | None:
        return None


__all__ = [
    "S",
    "T",
    "U",
    "E",
    "Ctx",
    "HookPhase",
    "HookPhases",
    "VALID_HOOK_PHASES",
    "StepFn",
    "HookPredicate",
    "AtomResult",
    "BaseCtx",
    "BootCtx",
    "IngressCtx",
    "RoutedCtx",
    "BoundCtx",
    "PlannedCtx",
    "GuardedCtx",
    "ExecutingCtx",
    "ResolvedCtx",
    "OperatedCtx",
    "EncodedCtx",
    "EmittingCtx",
    "EgressedCtx",
    "FailedCtx",
    "AtomFailure",
    "Atom",
    "StandardAtom",
    "promote",
    "fail",
    "has_error",
    "DependencyLike",
    "ResponseLike",
    "is_dependency_like",
    "is_response_like",
]



from __future__ import annotations

from typing import Final, Tuple, Type


class Boot: ...


class Ingress: ...


class Routed: ...


class Bound: ...


class Planned: ...


class Guarded: ...


class Executing: ...


class Resolved: ...


class Operated: ...


class Encoded: ...


class Emitting: ...


class Egressed: ...


class Failed: ...


Stage = Type[object]

STAGES: Final[Tuple[Stage, ...]] = (
    Boot,
    Ingress,
    Routed,
    Bound,
    Planned,
    Guarded,
    Executing,
    Resolved,
    Operated,
    Encoded,
    Emitting,
    Egressed,
    Failed,
)

_STAGE_ORDINAL: Final[dict[Stage, int]] = {
    stage: idx for idx, stage in enumerate(STAGES)
}


def stage_name(stage: Stage) -> str:
    return stage.__name__


def stage_ordinal(stage: Stage) -> int:
    try:
        return _STAGE_ORDINAL[stage]
    except KeyError as e:
        raise ValueError(f"unknown stage: {stage!r}") from e


def is_valid_stage(stage: object) -> bool:
    return stage in _STAGE_ORDINAL


def is_monotonic_transition(src: Stage, dst: Stage) -> bool:
    if src is Failed or dst is Failed:
        return True
    return stage_ordinal(src) <= stage_ordinal(dst)


__all__ = [
    "Boot",
    "Ingress",
    "Routed",
    "Bound",
    "Planned",
    "Guarded",
    "Executing",
    "Resolved",
    "Operated",
    "Encoded",
    "Emitting",
    "Egressed",
    "Failed",
    "Stage",
    "STAGES",
    "stage_name",
    "stage_ordinal",
    "is_valid_stage",
    "is_monotonic_transition",
]



from __future__ import annotations

from ...types import Atom, Ctx, EgressedCtx
from ...stages import Emitting, Egressed

import json
from typing import Any, Mapping

from ... import events as _ev
from ...types import ResponseLike, is_response_like

ANCHOR = _ev.EGRESS_ASGI_SEND


NO_BODY_STATUS = set(range(100, 200)) | {204, 205, 304}


def _json_default(value: Any) -> Any:
    model_dump = getattr(value, "model_dump", None)
    if callable(model_dump):
        return model_dump()

    table = getattr(value, "__table__", None)
    columns = getattr(table, "columns", None)
    if columns is not None:
        serialized: dict[str, Any] = {}
        for column in columns:
            key = getattr(column, "key", None)
            if isinstance(key, str) and key:
                serialized[key] = getattr(value, key, None)
        if serialized:
            return serialized

    obj_dict = getattr(value, "__dict__", None)
    if isinstance(obj_dict, dict):
        return {
            key: val
            for key, val in obj_dict.items()
            if not key.startswith("_") and not callable(val)
        }

    return str(value)


def finalize_transport_response(
    scope: Mapping[str, Any],
    status: int,
    headers: list[tuple[bytes, bytes]],
    body: bytes,
) -> tuple[list[tuple[bytes, bytes]], bytes]:
    method = str(scope.get("method", "GET")).upper()

    if method == "HEAD" or status in NO_BODY_STATUS:
        drop = {b"content-length", b"content-type", b"transfer-encoding"}
        headers = [(k, v) for (k, v) in headers if k.lower() not in drop]
        return headers, b""

    headers = [(k, v) for (k, v) in headers if k.lower() != b"content-length"]
    headers.append((b"content-length", str(len(body)).encode("latin-1")))
    return headers, body


async def _send_json(env: Any, status: int, payload: Any) -> None:
    scope = getattr(env, "scope", {}) or {}
    path = scope.get("path") if isinstance(scope, Mapping) else None
    is_jsonrpc = False
    if isinstance(path, str):
        normalized = path.rstrip("/") or "/"
        if normalized == "/rpc":
            is_jsonrpc = True
    if isinstance(payload, Mapping) and payload.get("jsonrpc") == "2.0":
        is_jsonrpc = True
    if is_jsonrpc and isinstance(payload, Mapping) and "error" not in payload:
        from tigrbl_typing.status.mappings import ERROR_MESSAGES, _HTTP_TO_RPC

        rpc_code = _HTTP_TO_RPC.get(int(status), -32603)
        detail = payload.get("detail")
        rpc_payload: dict[str, Any] = {
            "jsonrpc": "2.0",
            "error": {
                "code": rpc_code,
                "message": ERROR_MESSAGES.get(rpc_code, "Internal error"),
            },
            "id": None,
        }
        if isinstance(detail, Mapping):
            rpc_payload["error"]["data"] = dict(detail)
        elif detail is not None:
            rpc_payload["error"]["data"] = {"detail": detail}
        payload = rpc_payload
        status = 200

    body = json.dumps(
        payload,
        separators=(",", ":"),
        ensure_ascii=False,
        default=_json_default,
    ).encode("utf-8")
    headers = [(b"content-type", b"application/json")]
    headers, body = finalize_transport_response(scope, status, headers, body)
    await env.send(
        {
            "type": "http.response.start",
            "status": status,
            "headers": headers,
        }
    )
    await env.send(
        {
            "type": "http.response.body",
            "body": body,
            "more_body": False,
        }
    )


async def _send_transport_response(env: Any, ctx: Any) -> None:
    temp = getattr(ctx, "temp", None)
    if not isinstance(temp, dict):
        temp = {}
        setattr(ctx, "temp", temp)
    egress = temp.setdefault("egress", {})
    if not isinstance(egress, dict) or egress.get("response_sent"):
        return

    transport = egress.get("transport_response")
    if not isinstance(transport, dict):
        transport = {
            "status_code": int(getattr(ctx, "status_code", 200) or 200),
            "headers": {},
            "body": getattr(ctx, "result", None),
        }

    status = int(transport.get("status_code", getattr(ctx, "status_code", 200) or 200))
    headers_obj = transport.get("headers", {})
    headers: list[tuple[bytes, bytes]] = []
    if isinstance(headers_obj, dict):
        headers = [
            (str(k).encode("latin-1"), str(v).encode("latin-1"))
            for k, v in headers_obj.items()
            if v is not None
        ]

    body_obj = transport.get("body", None)
    if body_obj is None:
        body_obj = getattr(ctx, "result", None)
        if body_obj is None and status == 200:
            status = 204

    if isinstance(body_obj, Mapping) and set(body_obj.keys()) == {"body"}:
        nested = body_obj.get("body")
        if isinstance(nested, (bytes, bytearray)):
            body_obj = nested

    if not isinstance(
        body_obj, (bytes, bytearray, str, Mapping, list, tuple, set)
    ) and hasattr(body_obj, "body"):
        body_obj = getattr(body_obj, "body")

    if (
        isinstance(body_obj, Mapping)
        and "body" in body_obj
        and "status_code" not in body_obj
        and "headers" not in body_obj
    ):
        body_obj = body_obj.get("body")

    if (
        isinstance(body_obj, str)
        and body_obj.startswith("b'")
        and body_obj.endswith("'")
    ):
        try:
            body_obj = body_obj[2:-1].encode("latin-1", errors="ignore")
        except Exception:
            pass

    if isinstance(body_obj, (bytes, bytearray)):
        body = bytes(body_obj)
    elif body_obj is None:
        body = b""
    else:
        body = json.dumps(
            body_obj,
            separators=(",", ":"),
            default=_json_default,
        ).encode("utf-8")
        if not any(k.lower() == b"content-type" for k, _ in headers):
            headers.append((b"content-type", b"application/json; charset=utf-8"))

    headers, body = finalize_transport_response(env.scope, status, headers, body)

    await env.send(
        {"type": "http.response.start", "status": status, "headers": headers}
    )
    await env.send({"type": "http.response.body", "body": body, "more_body": False})
    egress["response_sent"] = True


def _headers_to_asgi(headers: Mapping[str, Any]) -> list[tuple[bytes, bytes]]:
    out: list[tuple[bytes, bytes]] = []
    for k, v in (headers or {}).items():
        if v is None:
            continue
        out.append((str(k).encode("latin-1"), str(v).encode("latin-1")))
    return out


def _json_bytes(obj: Any) -> bytes:
    return json.dumps(
        obj,
        separators=(",", ":"),
        ensure_ascii=False,
        default=_json_default,
    ).encode("utf-8")


def _to_headers_mapping(headers: Any) -> dict[str, Any]:
    if isinstance(headers, dict):
        return dict(headers)
    data = getattr(headers, "__dict__", None)
    if isinstance(data, dict):
        return dict(data)
    try:
        return dict(headers)
    except Exception:
        return {}


async def _run(obj: object | None, ctx: Any) -> None:
    del obj
    raw = getattr(ctx, "raw", None)
    send = getattr(raw, "send", None) if raw is not None else None
    scope = getattr(raw, "scope", None) if raw is not None else None
    if not callable(send) or not isinstance(scope, dict) or scope.get("type") != "http":
        return

    temp = getattr(ctx, "temp", None)
    egress = temp.get("egress") if isinstance(temp, dict) else None
    if isinstance(egress, dict) and egress.get("suppress_asgi_send"):
        has_transport_response = "transport_response" in egress
        if not has_transport_response:
            return

    resp = None
    resp_ns = getattr(ctx, "response", None)
    if resp_ns is not None:
        candidate = getattr(resp_ns, "result", None)
        if isinstance(candidate, ResponseLike) and is_response_like(candidate):
            resp = candidate

    temp = getattr(ctx, "temp", None)
    if not isinstance(temp, dict):
        temp = {}
        setattr(ctx, "temp", temp)
    egress = temp.setdefault("egress", {})
    if egress.get("response_sent"):
        return

    if resp is None:
        tr = egress.get("transport_response") if isinstance(egress, dict) else None

        status = int(tr.get("status_code", 200)) if isinstance(tr, dict) else 200
        headers = tr.get("headers", {}) if isinstance(tr, dict) else {}
        body_obj = tr.get("body", None) if isinstance(tr, dict) else None
        if body_obj is None:
            body_obj = getattr(ctx, "result", None)
            if body_obj is None and status == 200:
                status = 204

        if isinstance(body_obj, (bytes, bytearray)):
            body = bytes(body_obj)
        elif body_obj is None:
            body = b""
        else:
            body = _json_bytes(body_obj)
            headers = _to_headers_mapping(headers)
            headers.setdefault("content-type", "application/json; charset=utf-8")

        egress["transport_response"] = {
            "status_code": status,
            "headers": _to_headers_mapping(headers),
            "body": body,
        }
        asgi_headers = _headers_to_asgi(headers)
        asgi_headers, body = finalize_transport_response(
            scope, status, asgi_headers, body
        )
        await send(
            {
                "type": "http.response.start",
                "status": status,
                "headers": asgi_headers,
            }
        )
        await send({"type": "http.response.body", "body": body, "more_body": False})
        egress["response_sent"] = True
        return

    egress["transport_response"] = {
        "status_code": int(resp.status_code),
        "headers": {
            k.decode("latin-1"): v.decode("latin-1") for k, v in resp.raw_headers
        },
        "body": resp.body or b"",
    }
    body = resp.body or b""
    headers = list(resp.raw_headers)
    headers, body = finalize_transport_response(
        scope, int(resp.status_code), headers, body
    )
    await send(
        {
            "type": "http.response.start",
            "status": int(resp.status_code),
            "headers": headers,
        }
    )
    await send(
        {
            "type": "http.response.body",
            "body": body,
            "more_body": False,
        }
    )
    egress["response_sent"] = True


class AtomImpl(Atom[Emitting, Egressed]):
    name = "egress.asgi_send"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Emitting]) -> Ctx[Egressed]:
        await _run(obj, ctx)
        return ctx.promote(EgressedCtx)


INSTANCE = AtomImpl()

__all__ = [
    "ANCHOR",
    "NO_BODY_STATUS",
    "finalize_transport_response",
    "_send_json",
    "_send_transport_response",
    "INSTANCE",
]



from __future__ import annotations

import json
import uuid
from typing import Any, Mapping

from ... import events as _ev
from ...stages import Bound
from ...types import Atom, BoundCtx, Ctx

ANCHOR = _ev.DISPATCH_BINDING_PARSE


def _dispatch_dict(ctx: Any) -> dict[str, object]:
    temp = getattr(ctx, "temp", None)
    if not isinstance(temp, dict):
        temp = {}
        setattr(ctx, "temp", temp)
    obj = temp.setdefault("dispatch", {})
    if isinstance(obj, dict):
        return obj
    temp["dispatch"] = {}
    return temp["dispatch"]


def _run(obj: object | None, ctx: Any) -> None:
    del obj
    dispatch = _dispatch_dict(ctx)
    temp = getattr(ctx, "temp", None)
    route = temp.setdefault("route", {}) if isinstance(temp, dict) else {}
    protocol = str(dispatch.get("binding_protocol", "") or "")
    body = getattr(ctx, "body", None)

    if isinstance(body, (bytes, bytearray)):
        try:
            body = json.loads(bytes(body).decode("utf-8"))
        except Exception:
            body = None

    if not protocol and isinstance(body, Mapping) and body.get("jsonrpc") == "2.0":
        dispatch.setdefault("binding_protocol", "http.jsonrpc")
        dispatch["binding_selector"] = body.get("method")
        dispatch["rpc"] = dict(body)
        dispatch["rpc_method"] = body.get("method")
        dispatch["parsed_payload"] = body.get("params", {})
        if isinstance(route, dict):
            route["protocol"] = dispatch.get("binding_protocol")
            route["selector"] = dispatch.get("binding_selector")
            route["rpc_envelope"] = dict(body)
            route["payload"] = dispatch["parsed_payload"]
        return

    if not protocol and isinstance(body, list):
        is_jsonrpc_batch = all(
            isinstance(item, Mapping) and item.get("jsonrpc") == "2.0" for item in body
        )
        if is_jsonrpc_batch:
            dispatch["rpc_batch"] = [dict(item) for item in body]
            dispatch["parsed_payload"] = dispatch["rpc_batch"]
            if isinstance(route, dict):
                route["payload"] = dispatch["rpc_batch"]
            return

    if protocol.endswith(".jsonrpc"):
        if isinstance(body, Mapping):
            dispatch.setdefault("binding_selector", body.get("method"))
            dispatch["rpc"] = dict(body)
            dispatch["rpc_method"] = body.get("method")
            dispatch["parsed_payload"] = body.get("params", {})
            if isinstance(route, dict):
                route["selector"] = dispatch.get("binding_selector")
                route["rpc_envelope"] = dict(body)
        elif isinstance(body, list):
            dispatch["rpc_batch"] = [
                dict(item) if isinstance(item, Mapping) else item for item in body
            ]
            dispatch["parsed_payload"] = dispatch["rpc_batch"]
        if isinstance(route, dict):
            route["payload"] = body
    elif protocol.endswith(".rest"):
        if isinstance(body, (bytes, bytearray)):
            try:
                body = json.loads(bytes(body).decode("utf-8"))
            except Exception:
                body = None
        payload: dict[str, object] = {}
        query = getattr(ctx, "query", None)
        if isinstance(query, Mapping):
            for k, v in query.items():
                # parse_qs returns lists; unwrap single-element lists
                if isinstance(v, list) and len(v) == 1:
                    payload[str(k)] = v[0]
                else:
                    payload[str(k)] = v
        path_params = dispatch.get("path_params")
        if isinstance(path_params, Mapping):
            payload.update({str(k): v for k, v in path_params.items()})
        if isinstance(body, list):
            # Bulk operation: inject path params into each item
            coerced_params = {}
            if isinstance(path_params, Mapping):
                for pk, pv in path_params.items():
                    if isinstance(pv, str):
                        try:
                            coerced_params[pk] = uuid.UUID(pv)
                        except (ValueError, AttributeError):
                            coerced_params[pk] = pv
                    else:
                        coerced_params[pk] = pv
            for idx, item in enumerate(body):
                if isinstance(item, Mapping):
                    merged = dict(coerced_params)
                    merged.update(item)
                    body[idx] = merged
            dispatch["parsed_payload"] = body
            if isinstance(route, dict):
                route["payload"] = body
                route["path_params"] = (
                    dict(path_params) if isinstance(path_params, Mapping) else {}
                )
        elif isinstance(body, Mapping):
            payload.update({str(k): v for k, v in body.items()})
            dispatch["parsed_payload"] = payload
            if isinstance(route, dict):
                route["payload"] = payload
        else:
            dispatch["parsed_payload"] = payload
            if isinstance(route, dict):
                route["payload"] = payload
    else:
        dispatch["parsed_payload"] = body
        if isinstance(route, dict):
            route["payload"] = body


class AtomImpl(Atom[Bound, Bound]):
    name = "dispatch.binding_parse"
    anchor = ANCHOR

    @staticmethod
    async def _execute_rpc_batch(
        ctx: Any, batch_payload: list[Any]
    ) -> list[dict[str, Any]]:
        router_or_app = getattr(ctx, "router", None) or getattr(ctx, "app", None)
        request = getattr(ctx, "request", None)
        rpc_call = getattr(router_or_app, "rpc_call", None)
        responses: list[dict[str, Any]] = []

        for item in batch_payload:
            if not isinstance(item, Mapping):
                responses.append(
                    {
                        "jsonrpc": "2.0",
                        "error": {"code": -32600, "message": "Invalid Request"},
                        "id": None,
                    }
                )
                continue

            rpc_id = item.get("id")
            method = item.get("method")
            if not isinstance(method, str) or "." not in method:
                responses.append(
                    {
                        "jsonrpc": "2.0",
                        "error": {"code": -32601, "message": "Method not found"},
                        "id": rpc_id,
                    }
                )
                continue

            if not callable(rpc_call):
                responses.append(
                    {
                        "jsonrpc": "2.0",
                        "error": {
                            "code": -32603,
                            "message": "Internal error",
                            "data": {"detail": "RPC invoker unavailable."},
                        },
                        "id": rpc_id,
                    }
                )
                continue

            model_name, op_alias = method.split(".", 1)
            params = item.get("params", {})
            if params is None:
                params = {}

            try:
                result = await rpc_call(
                    model_name,
                    op_alias,
                    params,
                    request=request,
                    ctx={},
                )
            except AttributeError:
                responses.append(
                    {
                        "jsonrpc": "2.0",
                        "error": {"code": -32601, "message": "Method not found"},
                        "id": rpc_id,
                    }
                )
                continue
            except Exception as exc:
                responses.append(
                    {
                        "jsonrpc": "2.0",
                        "error": {
                            "code": -32603,
                            "message": "Internal error",
                            "data": {"detail": str(exc)},
                        },
                        "id": rpc_id,
                    }
                )
                continue

            responses.append({"jsonrpc": "2.0", "result": result, "id": rpc_id})

        return responses

    async def __call__(self, obj: object | None, ctx: Ctx[Bound]) -> Ctx[Bound]:
        _run(obj, ctx)
        temp = getattr(ctx, "temp", None)
        dispatch = temp.get("dispatch") if isinstance(temp, dict) else None
        rpc_batch = dispatch.get("rpc_batch") if isinstance(dispatch, dict) else None
        if isinstance(rpc_batch, list):
            egress = temp.setdefault("egress", {}) if isinstance(temp, dict) else {}
            if isinstance(egress, dict):
                egress["transport_response"] = {
                    "status_code": 200,
                    "body": await self._execute_rpc_batch(ctx, rpc_batch),
                }
        return ctx.promote(BoundCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



# pkgs/standards/tigrbl_atoms/tigrbl/atoms/wire/validate_in.py
from __future__ import annotations

from ...types import Atom, Ctx, ExecutingCtx
from ...stages import Executing

import datetime as _dt
import decimal as _dc
import uuid as _uuid
import logging
from typing import Any, Dict, Mapping, Optional, Tuple

from tigrbl_typing.status.exceptions import HTTPException
from tigrbl_typing.status.mappings import status as _status

from ... import events as _ev
from ..._opview_helpers import _ensure_ov, _ensure_schema_in, _ensure_temp

# PRE_HANDLER, runs after wire:build_in
ANCHOR = _ev.IN_VALIDATE  # "in:validate"

logger = logging.getLogger("uvicorn")


def _run(obj: Optional[object], ctx: Any) -> None:
    """
    wire:validate_in@in:validate

    Validates the normalized inbound payload (ctx.temp["in_values"]) using the
    schema collected by schema:collect_in plus tolerant signals from ColumnSpec.

    What it checks
    --------------
    - Required fields are present (ABSENT → error; present(None) is a separate check)
    - Nullability (value is None while column is non-nullable → error)
    - Lightweight type conformance (optional coercion if enabled)
    - String max_length (when available)
    - Author validators (validate_in / in_validator / validator_in on ColumnSpec/FieldSpec/IOSpec)

    Effects
    -------
    - Mutates ctx.temp["in_values"] with any successful coercions/validator returns
    - Writes diagnostics to:
        ctx.temp["in_errors"] : list of {field, code, message}
        ctx.temp["in_invalid"] : bool
        ctx.temp["in_coerced"] : tuple of coerced field names (if any)
    - Raises HTTPException(422) if any validation errors are found
    """
    logger.debug("Running wire:validate_in")
    temp = _ensure_temp(ctx)
    if isinstance(temp.get("rpc_error"), dict):
        return
    schema_in = _ensure_schema_in(ctx)

    in_values: Dict[str, Any] = dict(temp.get("in_values") or {})
    by_field: Mapping[str, Mapping[str, Any]] = schema_in.get("by_field", {})  # type: ignore[assignment]
    required: Tuple[str, ...] = tuple(schema_in.get("required", ()))  # type: ignore[assignment]

    errors: list[Dict[str, Any]] = []
    coerced: list[str] = []

    # 1) Required presence (ABSENT → error)
    for name in required:
        if name not in in_values:
            logger.debug("Required field %s missing", name)
            errors.append(
                _err(name, "required", "Field is required but was not provided.")
            )

    # 2) Per-field validation
    for name, value in list(in_values.items()):
        entry = by_field.get(name) or {}

        # Nullability
        nullable = entry.get("nullable", None)
        if value is None and nullable is False:
            logger.debug("Field %s is null but not nullable", name)
            errors.append(
                _err(name, "null_not_allowed", "Null is not allowed for this field.")
            )
            continue  # skip further checks for this field

        # Type (optionally coerce)
        target_type = entry.get("py_type")
        if value is not None and isinstance(target_type, type):
            allow_coerce = bool(entry.get("coerce", True))
            ok, new_val, msg = _coerce_if_needed(value, target_type, allow=allow_coerce)
            if not ok:
                logger.debug("Type mismatch for field %s", name)
                errors.append(
                    _err(
                        name,
                        "type_mismatch",
                        msg or f"Expected {_type_name(target_type)}.",
                    )
                )
                continue
            if new_val is not value:
                in_values[name] = new_val
                coerced.append(name)

        # Max length (strings)
        max_len = entry.get("max_length", None)
        if (
            isinstance(max_len, int)
            and max_len > 0
            and isinstance(in_values.get(name), str)
        ):
            if len(in_values[name]) > max_len:
                logger.debug("Field %s exceeds max_length %d", name, max_len)
                errors.append(
                    _err(name, "max_length", f"String exceeds max_length={max_len}.")
                )
                continue

        # Author-supplied validator(s)
        vfn = entry.get("validator")
        if callable(vfn) and in_values.get(name) is not None:
            try:
                out = vfn(in_values[name], ctx)
                if out is not None:
                    in_values[name] = out
            except Exception as e:
                logger.debug("Validator failed for field %s: %s", name, e)
                errors.append(
                    _err(name, "validator_failed", f"{type(e).__name__}: {e}")
                )
                continue

    # 3) Unknown/reserved key policy (handled after build_in captured samples)
    unknown = tuple(temp.get("in_unknown") or ())
    if unknown:
        reserved = _reserved_input_keys(ctx, by_field=by_field)
        for k in unknown:
            if k in reserved:
                errors.append(
                    _err(k, "reserved_input", "Field is not accepted on input.")
                )

    if unknown and _reject_unknown(ctx):
        logger.debug("Rejecting unknown fields: %s", unknown)
        for k in unknown:
            errors.append(_err(k, "unknown_field", "Unknown input key."))

    # Persist results + diagnostics
    temp["in_values"] = in_values
    if coerced:
        temp["in_coerced"] = tuple(coerced)

    if errors:
        temp["in_errors"] = errors
        temp["in_invalid"] = True
        logger.debug("Validation errors found: %s", errors)
        raise HTTPException(
            status_code=_status.HTTP_422_UNPROCESSABLE_ENTITY, detail=errors
        )
    else:
        temp["in_invalid"] = False
        logger.debug("Inbound payload validated successfully")


# ──────────────────────────────────────────────────────────────────────────────
# Internals
# ──────────────────────────────────────────────────────────────────────────────


def _reject_unknown(ctx: Any) -> bool:
    """
    Check cfg for a 'reject_unknown_fields' (bool); default False.
    """
    cfg = getattr(ctx, "cfg", None)
    val = getattr(cfg, "reject_unknown_fields", None)
    return bool(val) if isinstance(val, bool) else False


def _err(field: str, code: str, msg: str) -> Dict[str, Any]:
    return {"field": field, "code": code, "message": msg}


def _type_name(t: type) -> str:
    return getattr(t, "__name__", str(t))


# ── coercion helpers ──────────────────────────────────────────────────────────


def _coerce_if_needed(
    value: Any, target: type, *, allow: bool
) -> Tuple[bool, Any, Optional[str]]:
    """Return (ok, new_value, error_message). Only coerces when allow=True."""
    if isinstance(value, target):
        return True, value, None
    if not allow:
        return False, value, f"Expected {_type_name(target)}."
    try:
        coerced = _coerce(value, target)
        return True, coerced, None
    except Exception:
        return False, value, f"Could not convert to {_type_name(target)}."


def _coerce(value: Any, target: type) -> Any:
    if target is str:
        return str(value)
    if target is int:
        if isinstance(value, bool):
            return int(value)
        return int(str(value).strip())
    if target is float:
        return float(str(value).strip())
    if target is bool:
        if isinstance(value, bool):
            return value
        s = str(value).strip().lower()
        if s in {"true", "1", "yes", "y", "on"}:
            return True
        if s in {"false", "0", "no", "n", "off"}:
            return False
        raise ValueError("not a boolean")
    if target is _dc.Decimal:
        return _dc.Decimal(str(value).strip())
    if target is _uuid.UUID:
        return _uuid.UUID(str(value))
    if target is _dt.datetime:
        # allow both date-time and date-only (promote to midnight)
        s = str(value).strip()
        try:
            return _dt.datetime.fromisoformat(s)
        except Exception:
            d = _dt.date.fromisoformat(s)
            return _dt.datetime.combine(d, _dt.time())
    if target is _dt.date:
        return _dt.date.fromisoformat(str(value).strip())
    if target is _dt.time:
        return _dt.time.fromisoformat(str(value).strip())
    # Fallback: try direct construction
    return target(value)


def _reserved_input_keys(
    ctx: Any, *, by_field: Mapping[str, Mapping[str, Any]]
) -> set[str]:
    """Return keys that are reserved by paired-field IO plumbing."""
    reserved: set[str] = set()

    try:
        ov = _ensure_ov(ctx)
    except Exception:
        return reserved

    for field, desc in (getattr(ov, "paired_index", {}) or {}).items():
        if isinstance(field, str) and field and field not in by_field:
            reserved.add(field)
        alias = desc.get("alias") if isinstance(desc, Mapping) else None
        if isinstance(alias, str) and alias:
            reserved.add(alias)

    return reserved


class AtomImpl(Atom[Executing, Executing]):
    name = "wire.validate_in"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Executing]) -> Ctx[Executing]:
        _run(obj, ctx)
        return ctx.promote(ExecutingCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



from __future__ import annotations

from ...types import Atom, Ctx, EncodedCtx
from ...stages import Operated, Encoded
from typing import Any, Optional

from ... import events as _ev
from .templates import render_template
from .renderer import ResponseHints

ANCHOR = _ev.OUT_DUMP  # "out:dump"


async def _run(obj: Optional[object], ctx: Any) -> None:
    """response:template@out:dump

    Render a template if configured on ``ctx.response``.
    """
    resp_ns = getattr(ctx, "response", None)
    req = getattr(ctx, "request", None)
    if resp_ns is None or req is None:
        return
    tmpl = getattr(resp_ns, "template", None)
    if not tmpl:
        return
    result = getattr(resp_ns, "result", None)
    context = result if isinstance(result, dict) else {"data": result}
    html = await render_template(
        name=tmpl.name,
        context=context,
        search_paths=tmpl.search_paths,
        package=tmpl.package,
        auto_reload=bool(tmpl.auto_reload),
        filters=tmpl.filters,
        globals_=tmpl.globals,
        request=req,
    )
    resp_ns.result = html
    hints = getattr(resp_ns, "hints", None)
    if hints is None:
        hints = ResponseHints(status_code=int(getattr(ctx, "status_code", 200) or 200))
        resp_ns.hints = hints
    if not hints.media_type:
        hints.media_type = "text/html"


class AtomImpl(Atom[Operated, Encoded]):
    name = "response.template"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Operated]) -> Ctx[Encoded]:
        await _run(obj, ctx)
        return ctx.promote(EncodedCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



from __future__ import annotations

from ...types import Atom, Ctx, EncodedCtx
from ...stages import Encoded

from typing import Any, MutableMapping

from ... import events as _ev
from .._temp import _ensure_temp

ANCHOR = _ev.EGRESS_ENVELOPE_APPLY


def _is_jsonrpc(ctx: Any, egress: MutableMapping[str, Any]) -> bool:
    route = getattr(ctx, "gw_raw", None)
    if getattr(route, "kind", None) == "jsonrpc":
        return True

    path = getattr(route, "path", None)
    app = getattr(ctx, "app", None)
    prefix = getattr(app, "jsonrpc_prefix", None)
    if isinstance(path, str) and isinstance(prefix, str):
        if (path.rstrip("/") or "/") == (prefix.rstrip("/") or "/"):
            return True

    explicit = egress.get("response_kind")
    if explicit == "jsonrpc":
        return True

    route_temp = getattr(ctx, "temp", None)
    route_ns = route_temp.get("route") if isinstance(route_temp, dict) else None
    rpc_env = route_ns.get("rpc_envelope") if isinstance(route_ns, dict) else None
    if isinstance(rpc_env, dict) and rpc_env.get("jsonrpc") == "2.0":
        return True

    app = getattr(ctx, "app", None)
    prefix = getattr(app, "jsonrpc_prefix", None)
    raw = getattr(ctx, "raw", None)
    scope = getattr(raw, "scope", None)
    path = scope.get("path") if isinstance(scope, dict) else None
    if isinstance(prefix, str) and isinstance(path, str):
        norm_prefix = prefix.rstrip("/") or "/"
        norm_path = path.rstrip("/") or "/"
        if norm_path == norm_prefix:
            return True

    return False


def _run(obj: object | None, ctx: Any) -> None:
    del obj
    temp = _ensure_temp(ctx)
    egress = temp.setdefault("egress", {})

    payload = egress.get("wire_payload")
    if payload is None and "result" in egress:
        payload = egress.get("result")
    if payload is None and hasattr(ctx, "result"):
        payload = getattr(ctx, "result", None)
    temp = getattr(ctx, "temp", None)
    rpc_error = temp.get("rpc_error") if isinstance(temp, dict) else None
    if payload is None and not isinstance(rpc_error, dict):
        return

    if _is_jsonrpc(ctx, egress):
        request_rpc = getattr(getattr(ctx, "gw_raw", None), "rpc", None)
        if not isinstance(request_rpc, dict) and isinstance(temp, dict):
            route = temp.get("route")
            maybe_env = route.get("rpc_envelope") if isinstance(route, dict) else None
            if isinstance(maybe_env, dict):
                request_rpc = maybe_env

        rpc_id = request_rpc.get("id") if isinstance(request_rpc, dict) else None
        if isinstance(rpc_error, dict):
            egress["enveloped"] = {
                "jsonrpc": "2.0",
                "error": dict(rpc_error),
                "id": rpc_id,
            }
        else:
            egress["enveloped"] = {
                "jsonrpc": "2.0",
                "result": payload,
                "id": rpc_id,
            }
        return

    envelope = egress.get("envelope")
    if isinstance(envelope, dict):
        data = dict(envelope)
        data.setdefault("data", payload)
        egress["enveloped"] = data
    else:
        egress["enveloped"] = payload


class AtomImpl(Atom[Encoded, Encoded]):
    name = "egress.envelope_apply"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Encoded]) -> Ctx[Encoded]:
        _run(obj, ctx)
        return ctx.promote(EncodedCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



from __future__ import annotations

from collections import defaultdict
from typing import Any, Mapping, Sequence
from urllib.parse import parse_qs

from tigrbl_typing.gw.raw import GwRouteEnvelope

from ... import events as _ev
from ..._request import Request
from ...stages import Ingress
from ...types import Atom, Ctx, IngressCtx
from .._temp import _ensure_temp

ANCHOR = _ev.INGRESS_TRANSPORT_EXTRACT


def _decode_headers(raw_headers: object) -> dict[str, str]:
    if not isinstance(raw_headers, list):
        return {}
    headers: dict[str, str] = {}
    for item in raw_headers:
        if not (isinstance(item, tuple) and len(item) == 2):
            continue
        key, value = item
        if isinstance(key, (bytes, bytearray)) and isinstance(
            value, (bytes, bytearray)
        ):
            headers[bytes(key).decode("latin-1").lower()] = bytes(value).decode(
                "latin-1"
            )
    return headers


def _decode_headers_multi(raw_headers: object) -> dict[str, list[str]]:
    if not isinstance(raw_headers, list):
        return {}
    out: dict[str, list[str]] = defaultdict(list)
    for item in raw_headers:
        if not (isinstance(item, tuple) and len(item) == 2):
            continue
        key, value = item
        if isinstance(key, (bytes, bytearray)) and isinstance(
            value, (bytes, bytearray)
        ):
            out[bytes(key).decode("latin-1").lower()].append(
                bytes(value).decode("latin-1")
            )
    return dict(out)


def _normalize_query(query: object) -> dict[str, list[Any]]:
    if query is None:
        return {}
    if isinstance(query, Mapping):
        out: dict[str, list[Any]] = {}
        for key, value in query.items():
            if isinstance(value, Sequence) and not isinstance(
                value, (str, bytes, bytearray)
            ):
                out[str(key)] = list(value)
            else:
                out[str(key)] = [value]
        return out
    return {}


def _parse_query(raw_query: object) -> dict[str, list[str]]:
    if isinstance(raw_query, (bytes, bytearray)):
        return parse_qs(bytes(raw_query).decode("latin-1"), keep_blank_values=True)
    if isinstance(raw_query, str):
        return parse_qs(raw_query, keep_blank_values=True)
    return {}


async def _read_http_body(ctx: Any) -> object | None:
    body = getattr(ctx, "body", None)
    if body is not None:
        return body

    raw = getattr(ctx, "raw", None)
    receive = getattr(raw, "receive", None) if raw is not None else None
    scope = getattr(raw, "scope", None) if raw is not None else None
    if not (
        callable(receive) and isinstance(scope, dict) and scope.get("type") == "http"
    ):
        return None

    chunks: list[bytes] = []
    while True:
        message = await receive()
        if not isinstance(message, dict) or message.get("type") != "http.request":
            break
        chunk = message.get("body", b"")
        if isinstance(chunk, (bytes, bytearray)):
            chunks.append(bytes(chunk))
        if not bool(message.get("more_body", False)):
            break
    return b"".join(chunks)


async def _run(obj: object | None, ctx: Any) -> None:
    del obj

    raw = getattr(ctx, "raw", None)
    scope = getattr(raw, "scope", None) if raw is not None else None
    if not isinstance(scope, dict):
        return

    request = getattr(ctx, "request", None)
    if request is None:
        request = Request(scope, app=getattr(ctx, "app", None))
        setattr(ctx, "request", request)

    method = str(scope.get("method", getattr(request, "method", "") or "")).upper()
    path = str(scope.get("path", getattr(request, "path", "/") or "/"))
    query = _parse_query(scope.get("query_string", b"")) or _normalize_query(
        getattr(request, "query_params", None)
    )
    headers = dict(getattr(request, "headers", {}) or {}) or _decode_headers_multi(
        scope.get("headers")
    )

    body = await _read_http_body(ctx)
    if body is None:
        gw_raw = getattr(ctx, "gw_raw", None)
        body = getattr(gw_raw, "body", None) if gw_raw is not None else None

    if isinstance(body, bytearray):
        body = bytes(body)
    if isinstance(body, memoryview):
        body = body.tobytes()
    if isinstance(body, str):
        body = body.encode("utf-8")

    transport = "ws" if scope.get("type") == "websocket" else "http"
    scheme = str(scope.get("scheme", "http")).lower()

    gw_raw = GwRouteEnvelope(
        transport=transport,
        scheme="wss"
        if scheme in {"wss", "https"} and transport == "ws"
        else (
            "https" if scheme == "https" else ("ws" if transport == "ws" else "http")
        ),
        kind="unknown",
        method=method if transport == "http" else None,
        path=path,
        headers=_decode_headers(scope.get("headers")),
        query={str(k): [str(vv) for vv in values] for k, values in query.items()},
        body=body if transport == "http" else None,
        ws_event=getattr(ctx, "raw_event", None) if transport == "ws" else None,
        rpc=None,
    )
    setattr(ctx, "gw_raw", gw_raw)

    temp = _ensure_temp(ctx)
    ingress = temp.setdefault("ingress", {})
    ingress.update(
        {
            "transport": transport,
            "scheme": gw_raw.scheme,
            "method": method,
            "path": path,
            "headers": headers,
            "query": query,
            "body": body,
            "request": request,
        }
    )
    setattr(ctx, "method", method)
    setattr(ctx, "path", path)
    setattr(ctx, "headers", headers)
    setattr(ctx, "query", query)
    if body is not None:
        setattr(ctx, "body", body)


class AtomImpl(Atom[Ingress, Ingress]):
    name = "ingress.transport_extract"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Ingress]) -> Ctx[Ingress]:
        await _run(obj, ctx)
        return ctx.promote(IngressCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



from __future__ import annotations

from typing import Annotated, Any, Iterable, get_args, get_origin

from tigrbl_ops_oltp.crud.params import Param


def _request_field(req: Any, name: str) -> Any:
    if isinstance(req, dict):
        return req.get(name)
    return getattr(req, name, None)


def split_annotated(annotation: Any) -> tuple[Any, tuple[Any, ...]]:
    if get_origin(annotation) is Annotated:
        args = get_args(annotation)
        if args:
            return args[0], tuple(args[1:])
    return annotation, ()


def annotation_marker(extras: Iterable[Any], marker_type: type[Any]) -> Any | None:
    for extra in extras:
        if isinstance(extra, marker_type):
            return extra
    return None


def is_request_annotation(annotation: Any) -> bool:
    if isinstance(annotation, type) and annotation.__name__ == "Request":
        return True
    if isinstance(annotation, str):
        normalized = annotation.strip().strip("\"'")
        return normalized.rsplit(".", maxsplit=1)[-1] == "Request"
    return False


def _load_body(req: Any) -> Any:
    if isinstance(req, dict):
        body = req.get("body", req.get("payload"))
        if body is not None:
            return body
    json_sync = getattr(req, "json_sync", None)
    if callable(json_sync):
        return json_sync()
    return None


def extract_param_value(
    marker: Param,
    req: Any,
    param_name: str,
    body_cache: Any | None,
) -> tuple[Any, bool]:
    alias = marker.alias or param_name
    query_params = _request_field(req, "query_params") or {}
    path_params = _request_field(req, "path_params") or {}
    headers = _request_field(req, "headers") or {}

    if marker.location == "query":
        if alias in query_params:
            return query_params[alias], True
        return None, False
    if marker.location == "path":
        if alias in path_params:
            return path_params[alias], True
        return None, False
    if marker.location == "header":
        key = alias.lower()
        if key in headers:
            return headers[key], True
        return None, False
    if marker.location == "body":
        if body_cache is None:
            body_cache = _load_body(req)
        if isinstance(body_cache, dict):
            if marker.alias is None and param_name == "body":
                return body_cache, True
            if alias in body_cache:
                return body_cache[alias], True
            if param_name in body_cache:
                return body_cache[param_name], True
            if marker.alias is None:
                return body_cache, True
        elif body_cache is not None:
            return body_cache, True
    return None, False


__all__ = [
    "Param",
    "annotation_marker",
    "extract_param_value",
    "is_request_annotation",
    "split_annotated",
]



from __future__ import annotations

from ...types import Atom, Ctx, EncodedCtx
from ...stages import Encoded

from typing import Any

from ... import events as _ev
from .._temp import _ensure_temp, _response_payload

ANCHOR = _ev.EGRESS_OUT_DUMP


def _run(obj: object | None, ctx: Any) -> None:
    del obj
    temp = _ensure_temp(ctx)
    egress = temp.setdefault("egress", {})
    if "wire_payload" in egress:
        return

    wire_payload = _response_payload(ctx)
    if wire_payload is None:
        wire_payload = egress.get("result", getattr(ctx, "result", None))

    if wire_payload is not None:
        egress["wire_payload"] = wire_payload


class AtomImpl(Atom[Encoded, Encoded]):
    name = "egress.out_dump"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Encoded]) -> Ctx[Encoded]:
        _run(obj, ctx)
        return ctx.promote(EncodedCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



from __future__ import annotations

from ...types import Atom, Ctx, GuardedCtx
from ...stages import Planned, Guarded

import inspect
from typing import Any, Callable

from ._param_resolver import (
    Param,
    annotation_marker,
    extract_param_value,
    is_request_annotation,
    split_annotated,
)
from tigrbl_typing.status.exceptions import HTTPException
from tigrbl_typing.status.mappings import status
from ...types import DependencyLike, is_dependency_like
from ... import events as _ev

ANCHOR = _ev.DEP_SECURITY


async def invoke_dependency(router: Any, dep: Callable[..., Any], req: Any) -> Any:
    provider = getattr(router, "dependency_overrides_provider", None) or router
    overrides = getattr(provider, "dependency_overrides", {})
    dep = overrides.get(dep, dep)
    kwargs: dict[str, Any] = {}

    for name, param in inspect.signature(dep).parameters.items():
        base_annotation, extras = split_annotated(param.annotation)
        dependency_marker = annotation_marker(extras, DependencyLike)
        param_marker = annotation_marker(extras, Param)

        if (
            is_request_annotation(base_annotation)
            or name == "request"
            or name.lower().endswith("request")
        ):
            kwargs[name] = req
        elif is_dependency_like(param.default) or dependency_marker is not None:
            child = (
                param.default.dependency
                if is_dependency_like(param.default)
                else dependency_marker.dependency
            )
            kwargs[name] = await invoke_dependency(router, child, req)
        elif isinstance(param.default, Param) or param_marker is not None:
            marker = param.default if isinstance(param.default, Param) else param_marker
            value, found = extract_param_value(marker, req, name, None)
            if found:
                kwargs[name] = value
            elif marker.required:
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail=f"Missing required parameter: {marker.alias or name}",
                )
            else:
                kwargs[name] = marker.default
        elif name in req.path_params:
            kwargs[name] = req.path_params[name]
        elif name in req.query_params:
            kwargs[name] = req.query_params[name]
        elif param.default is not inspect._empty:
            kwargs[name] = param.default

    out = dep(**kwargs)
    if inspect.isgenerator(out):
        try:
            value = next(out)
        except StopIteration:
            return None
        cleanups = getattr(req.state, "_dependency_cleanups", None)
        if isinstance(cleanups, list):
            cleanups.append(out.close)
        return value

    if inspect.isasyncgen(out):
        try:
            value = await anext(out)
        except StopAsyncIteration:
            return None
        cleanups = getattr(req.state, "_dependency_cleanups", None)
        if isinstance(cleanups, list):
            cleanups.append(out.aclose)
        return value

    if inspect.isawaitable(out):
        return await out
    return out


async def _run(dep: object | None, ctx: Any) -> Any:
    fn = getattr(dep, "dependency", dep)
    if not callable(fn):
        return None

    ctx_map = ctx if isinstance(ctx, dict) else {}
    req = getattr(ctx, "request", None) or ctx_map.get("request") or ctx_map.get("req")
    router = getattr(ctx, "router", None) or ctx_map.get("router")

    if req is not None and router is not None:
        value = await invoke_dependency(router, fn, req)
        if getattr(fn, "__tigrbl_require_auth__", False) and value is None:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Unauthorized",
            )
        return value

    try:
        rv = fn(ctx)
    except TypeError:
        rv = fn()
    if inspect.isawaitable(rv):
        return await rv
    return rv


class AtomImpl(Atom[Planned, Guarded]):
    name = "dep.security"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Planned]) -> Ctx[Guarded]:
        await _run(obj, ctx)
        return ctx.promote(GuardedCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]



from __future__ import annotations

from typing import Any, Mapping

from ... import events as _ev
from ...stages import Ingress, Bound
from ...types import Atom, BoundCtx, Ctx

ANCHOR = _ev.DISPATCH_BINDING_MATCH


def _dispatch_dict(ctx: Any) -> dict[str, object]:
    temp = getattr(ctx, "temp", None)
    if not isinstance(temp, dict):
        temp = {}
        setattr(ctx, "temp", temp)
    obj = temp.setdefault("dispatch", {})
    if isinstance(obj, dict):
        return obj
    temp["dispatch"] = {}
    return temp["dispatch"]


def _method(ctx: Any) -> str:
    return str(getattr(ctx, "method", "") or "").upper()


def _path(ctx: Any) -> str:
    return str(getattr(ctx, "path", "/") or "/")


def _rpc_method(dispatch: Mapping[str, object], ctx: Any) -> str | None:
    rpc = dispatch.get("rpc")
    if isinstance(rpc, Mapping):
        method = rpc.get("method")
        if isinstance(method, str):
            return method
    body_json = dispatch.get("body_json")
    if isinstance(body_json, Mapping):
        method = body_json.get("method")
        if isinstance(method, str):
            return method
    temp = getattr(ctx, "temp", None)
    if isinstance(temp, Mapping):
        ingress = temp.get("ingress")
        if isinstance(ingress, Mapping):
            ingress_body_json = ingress.get("body_json")
            if isinstance(ingress_body_json, Mapping):
                method = ingress_body_json.get("method")
                if isinstance(method, str):
                    return method
    body = getattr(ctx, "body", None)
    if isinstance(body, (bytes, bytearray)):
        try:
            import json

            decoded = json.loads(bytes(body).decode("utf-8"))
        except Exception:
            decoded = None
        if isinstance(decoded, Mapping):
            method = decoded.get("method")
            if isinstance(method, str):
                return method
    if isinstance(body, Mapping):
        method = body.get("method")
        if isinstance(method, str):
            return method
    return None


def _run(obj: object | None, ctx: Any) -> None:
    del obj
    dispatch = _dispatch_dict(ctx)
    temp = getattr(ctx, "temp", None)
    route = temp.setdefault("route", {}) if isinstance(temp, dict) else {}
    plan = getattr(ctx, "kernel_plan", None) or getattr(ctx, "plan", None)
    proto_indices = getattr(plan, "proto_indices", {}) if plan is not None else {}

    method = _method(ctx)
    path = _path(ctx)
    rpc_method = _rpc_method(dispatch, ctx)

    matched_proto: str | None = None
    matched_selector: str | None = None
    matched_path_params: dict[str, object] = {}

    if isinstance(proto_indices, Mapping):
        for proto, bucket in proto_indices.items():
            if not isinstance(proto, str) or not isinstance(bucket, Mapping):
                continue
            if (
                proto.endswith(".jsonrpc")
                and isinstance(rpc_method, str)
                and rpc_method in bucket
            ):
                matched_proto = proto
                matched_selector = rpc_method
                break
            if proto.endswith(".rest"):
                exact = bucket.get("exact")
                selector = f"{method} {path}"
                if isinstance(exact, Mapping) and selector in exact:
                    matched_proto = proto
                    matched_selector = selector
                    break
                templated = bucket.get("templated")
                if isinstance(templated, list):
                    for entry in templated:
                        if not isinstance(entry, Mapping):
                            continue
                        pattern = entry.get("pattern")
                        if pattern is None or not hasattr(pattern, "match"):
                            continue
                        m = pattern.match(path)
                        if m is None:
                            continue
                        entry_method = str(entry.get("method", "") or "").upper()
                        if entry_method and entry_method != method:
                            continue
                        matched_proto = proto
                        matched_selector = str(entry.get("selector") or selector)
                        names = entry.get("names")
                        if isinstance(names, (list, tuple)):
                            matched_path_params = {
                                str(name): value
                                for name, value in zip(names, m.groups())
                            }
                        break
                    if matched_selector is not None:
                        break
            if proto in {"ws", "wss"}:
                exact = bucket.get("exact")
                if isinstance(exact, Mapping) and path in exact:
                    matched_proto = proto
                    matched_selector = path
                    break

    if matched_proto is not None:
        dispatch["binding_protocol"] = matched_proto
        if isinstance(route, dict):
            route["protocol"] = matched_proto
        setattr(ctx, "protocol", matched_proto)
    if matched_selector is not None:
        dispatch["binding_selector"] = matched_selector
        if isinstance(route, dict):
            route["selector"] = matched_selector
        setattr(ctx, "selector", matched_selector)
    if matched_path_params:
        dispatch["path_params"] = matched_path_params
        if isinstance(route, dict):
            route["path_params"] = matched_path_params
        setattr(ctx, "path_params", matched_path_params)


class AtomImpl(Atom[Ingress, Bound]):
    name = "dispatch.binding_match"
    anchor = ANCHOR

    async def __call__(self, obj: object | None, ctx: Ctx[Ingress]) -> Ctx[Bound]:
        _run(obj, ctx)
        return ctx.promote(BoundCtx)


INSTANCE = AtomImpl()

__all__ = ["ANCHOR", "INSTANCE"]
